# -*- coding: utf-8 -*-

from ..base import Base, Property
from ..types.state import State
from ..models.transition import TransitionModel


class Platform(Base):
    """Platform base class

    A platform represents a random object defined as a :class:`~.State`
    that moves according to a given :class:`~.TransitionModel`.
    """

    state = Property(State, doc="The platform state at any given point")
    transition_model = Property(
        TransitionModel, doc="Transition model")

    def move(self, timestamp=None, **kwargs):
        """Propagate the platform position using the :attr:`transition_model`.

        Parameters
        ----------
        timestamp: :class:`datetime.datetime`, optional
            A timestamp signifying when the end of the maneuver \
            (the default is `None`)

        Notes
        -----
        This methods updates the value of :attr:`position`.

        Any provided `kwargs` are forwarded to the :attr:`transition_model`.

        If :attr:`transition_model` or `timestamp` is `None`, the method has
        no effect, but will return successfully.

        """
        # Compute time_interval
        try:
            time_interval = timestamp - self.state.timestamp
        except TypeError:
            # TypeError: (timestamp or prior.timestamp) is None
            time_interval = None

        # Return without moving static platforms
        if self.transition_model is None:
            self.state.timestamp = timestamp
            return self

        self.state = State(
            state_vector=self.transition_model.function(
                state_vector=self.state.state_vector,
                timestamp=timestamp,
                time_interval=time_interval,
                **kwargs),
            timestamp=timestamp)



import json
from datetime import datetime

import numpy as np

from .base import DetectionReader
from .file import TextFileReader
from ..types.detection import Detection


class JSON_AISDetectionReader(DetectionReader, TextFileReader):
    """A simple detection reader for JSON files of AIS (maritime
    transponder) detections.

    This particular JSON AIS reader is written to read the JSON format
    used by files downloaded from 'www.aishub.net' (must be a contributor
    to their AIS database to download their files).


    JSON format::

        [{"ERROR": "false"},
        [{"NAME": "MARTCILINO", "MMSI": 205466990, "LONGITUDE": 3078401,
        "TIME": "1516233686", "LATITUDE": 31215032, ...},
        {"NAME": "MARTCILINO", "MMSI": 205466990, "LONGITUDE": 3064592,
        "TIME": "1516234275", "LATITUDE": 31227463, ...}]]

    Notes
    -----
    * TIME is in Linux Epoch format
    * LONGITUDE and LATITUDE are (long/lat degrees)*(600,000)
    * MMSI is unique ship identifier
    * The AIS detection attributes for lattitude, longitude, and timestamp are
      saved as the attributes of a 'Detection'; the other attributes are saved
      as the dictionary 'metadata' attribute of a 'Detection'.
    """

    # path - inherited from 'TextFileReader'->'FileReader'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._detections = set()

    @property
    def detections(self):
        return self._detections.copy()

    def detections_gen(self):

        # read data from the JSON file
        with self.path.open(encoding=self.encoding) as json_file:
            file_data = json.load(json_file)
            file_data = file_data[1]

            # sort the imported AIS Detection list by time of detection
            file_data.sort(key=lambda x: x['TIME'])

            for record in file_data:

                # extract latitude and longitude values from JSON record
                lat_value = float(record['LATITUDE'])/600000
                lon_value = float(record['LONGITUDE'])/600000

                # extract timestamp values from JSON record
                time_value = datetime.utcfromtimestamp(float(
                        record['TIME']))

                # delete lat, lon, timestamp from JSON record;
                # the rest is metadata
                del record['LATITUDE']
                del record['LONGITUDE']
                del record['TIME']

                # form Detection object from JSON record
                detect = Detection(
                            np.array([[lon_value], [lat_value]],
                                     dtype=np.float32),
                            time_value, metadata=record)
                self._detections = {detect}
                yield time_value, self.detections



# -*- coding: utf-8 -*-

from .base import MetricGenerator
from ..types.metric import TimeRangeMetric
from ..types.time import TimeRange


class BasicMetrics(MetricGenerator):
    """Calculates simple metrics like number of tracks, truth and
    ratio of track-to-truth"""

    def compute_metric(self, manager, *args, **kwargs):
        """Compute the metric using the data in the metric manager

        Parameters
        ----------
        manager : MetricManager
            containing the data to be used to create the metric(s)

        Returns
        ----------
        : list of :class:`~.Metric`
            Contains the metric information
        """

        metrics = []

        # Make a list of all the unique timestamps used
        timestamps = {state.timestamp for state in manager.tracks}
        timestamps |= {state.timestamp
                       for path in manager.groundtruth_paths
                       for state in path}

        # Number of tracks
        metrics.append(TimeRangeMetric(
            title='Number of targets',
            value=len(manager.groundtruth_paths),
            time_range=TimeRange(
                start_timestamp=min(timestamps),
                end_timestamp=max(timestamps)),
            generator=self))

        metrics.append(TimeRangeMetric(
            title='Number of tracks',
            value=len(manager.tracks),
            time_range=TimeRange(
                start_timestamp=min(timestamps),
                end_timestamp=max(timestamps)),
            generator=self))

        metrics.append(TimeRangeMetric(
            title='Track-to-target ratio',
            value=len(manager.tracks) / len(manager.groundtruth_paths),
            time_range=TimeRange(
                start_timestamp=min(timestamps),
                end_timestamp=max(timestamps)),
            generator=self))

        return metrics



# -*- coding: utf-8 -*-
import datetime
from collections.abc import MutableSequence

import numpy as np

from ..base import Property
from .array import StateVector, CovarianceMatrix
from .base import Type
from .particle import Particle


class State(Type):
    """State type.

    Most simple state type, which only has time and a state vector."""
    timestamp = Property(datetime.datetime, default=None,
                         doc="Timestamp of the state. Default None.")
    state_vector = Property(StateVector, doc='State vector.')

    def __init__(self, state_vector, *args, **kwargs):
        # Don't cast away subtype of state_vector if not necessary
        if state_vector is not None \
                and not isinstance(state_vector, StateVector):
            state_vector = StateVector(state_vector)
        super().__init__(state_vector, *args, **kwargs)

    @property
    def ndim(self):
        """The number of dimensions represented by the state."""
        return self.state_vector.shape[0]


class StateMutableSequence(Type, MutableSequence):
    """A mutable sequence for :class:`~.State` instances

    This sequence acts like a regular list object for States, as well as
    proxying state attributes to the last state in the sequence. This sequence
    can also be indexed/sliced by :class:`datetime.datetime` instances.

    Example
    -------
    >>> t0 = datetime.datetime(2018, 1, 1, 14, 00)
    >>> t1 = t0 + datetime.timedelta(minutes=1)
    >>> state0 = State([[0]], t0)
    >>> sequence = StateMutableSequence([state0])
    >>> print(sequence.state_vector, sequence.timestamp)
    [[0]] 2018-01-01 14:00:00
    >>> sequence.append(State([[1]], t1))
    >>> for state in sequence[t1:]:
    ...     print(state.state_vector, state.timestamp)
    [[1]] 2018-01-01 14:01:00
    """

    states = Property(
        [State],
        default=None,
        doc="The initial list of states. Default `None` which initialises"
            "with empty list.")

    def __init__(self, states=None, *args, **kwargs):
        if states is None:
            states = []
        super().__init__(states, *args, **kwargs)

    def __len__(self):
        return self.states.__len__()

    def __setitem__(self, index, value):
        return self.states.__setitem__(index, value)

    def __delitem__(self, index):
        return self.states.__delitem__(index)

    def __getitem__(self, index):
        if isinstance(index, slice) and (
                isinstance(index.start, datetime.datetime)
                or isinstance(index.stop, datetime.datetime)):
            items = []
            for state in self.states:
                try:
                    if index.start and state.timestamp < index.start:
                        continue
                    if index.stop and state.timestamp >= index.stop:
                        continue
                except TypeError as exc:
                    raise TypeError(
                        'both indices must be `datetime.datetime` objects for'
                        'time slice') from exc
                items.append(state)
            return StateMutableSequence(items[::index.step])
        elif isinstance(index, datetime.datetime):
            for state in self.states:
                if state.timestamp == index:
                    return state
            else:
                raise IndexError('timestamp not found in states')
        elif isinstance(index, slice):
            return StateMutableSequence(self.states.__getitem__(index))
        else:
            return self.states.__getitem__(index)

    def __getattr__(self, item):
        if item.startswith("_"):
            # Don't proxy special/private attributes to `state`
            raise AttributeError(
                "{!r} object has no attribute {!r}".format(
                    type(self).__name__, item))
        else:
            return getattr(self.state, item)

    def insert(self, index, value):
        return self.states.insert(index, value)

    @property
    def state(self):
        return self.states[-1]


class GaussianState(State):
    """Gaussian State type

    This is a simple Gaussian state object, which, as the name suggests,
    is described by a Gaussian state distribution.
    """
    covar = Property(CovarianceMatrix, doc='Covariance matrix of state.')

    def __init__(self, state_vector, covar, *args, **kwargs):
        covar = CovarianceMatrix(covar)
        super().__init__(state_vector, covar, *args, **kwargs)
        if self.state_vector.shape[0] != self.covar.shape[0]:
            raise ValueError(
                "state vector and covar should have same dimensions")

    @property
    def mean(self):
        """The state mean, equivalent to state vector"""
        return self.state_vector


class WeightedGaussianState(GaussianState):
    """Weighted Gaussian State Type

    Gaussian State object with an associated weight.  Used as components
    for a GaussianMixtureState.
    """
    weight = Property(float, default=0, doc="Weight of the Gaussian State.")


class ParticleState(Type):
    """Particle State type

    This is a particle state object which describes the state as a
    distribution of particles"""

    timestamp = Property(datetime.datetime, default=None,
                         doc="Timestamp of the state. Default None.")
    particles = Property([Particle],
                         doc='List of particles representing state')

    @property
    def mean(self):
        """The state mean, equivalent to state vector"""
        result = np.average([p.state_vector for p in self.particles], axis=0,
                            weights=[p.weight for p in self.particles])
        # Convert type as may have type of weights
        return result.astype(np.float, copy=False)

    @property
    def state_vector(self):
        """The mean value of the particle states"""
        return self.mean

    @property
    def covar(self):
        cov = np.cov(np.hstack([p.state_vector for p in self.particles]),
                     ddof=0, aweights=[p.weight for p in self.particles])
        # Fix one dimensional covariances being returned with zero dimension
        if not cov.shape:
            cov = cov.reshape(1, 1)
        return cov



# -*- coding: utf-8 -*-
import uuid

from ..base import Property
from .state import State, StateMutableSequence
from .update import Update


class Track(StateMutableSequence):
    """Track type

    A :class:`~.StateMutableSequence` representing a track.
    """

    states = Property(
        [State],
        default=None,
        doc="The initial states of the track. Default `None` which initialises"
            "with empty list.")

    id = Property(
        str,
        default=None,
        doc="The unique track ID")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.id is None:
            self.id = str(uuid.uuid4())

    @property
    def metadata(self):
        """Returns metadata associated with a track.

        Parameters
        ----------
        None

        Returns
        -------
        : :class:`dict` of variable size
            All metadata associate with this track.
        """

        metadata = {}

        for state in self:
            if isinstance(state, Update) \
                    and state.hypothesis.measurement.metadata is not None:
                metadata.update(state.hypothesis.measurement.metadata)

        return metadata



# -*- coding: utf-8 -*-
from functools import lru_cache

from .base import Predictor
from ..types.particle import Particle
from ..types.prediction import ParticleStatePrediction


class ParticlePredictor(Predictor):
    """ParticlePredictor class

    An implementation of a Particle Filter predictor.
    """

    @lru_cache()
    def predict(self, prior, control_input=None, timestamp=None, **kwargs):
        """Particle Filter prediction step

        Parameters
        ----------
        prior : :class:`~.ParticleState`
            A prior state object
        control_input : :class:`~.State`, optional
            The control input. It will only have an effect if
            :attr:`control_model` is not `None` (the default is `None`)
        timestamp: :class:`datetime.datetime`, optional
            A timestamp signifying when the prediction is performed
            (the default is `None`)

        Returns
        -------
        : :class:`~.ParticleStatePrediction`
            The predicted state
        """
        # Compute time_interval
        try:
            time_interval = timestamp - prior.timestamp
        except TypeError:
            # TypeError: (timestamp or prior.timestamp) is None
            time_interval = None

        new_particles = []
        for particle in prior.particles:
            new_state_vector = self.transition_model.function(
                particle.state_vector,
                time_interval=time_interval,
                **kwargs)
            new_particles.append(
                Particle(new_state_vector,
                         weight=particle.weight,
                         parent=particle.parent))

        return ParticleStatePrediction(new_particles, timestamp=timestamp)



# -*- coding: utf-8 -*-
from math import log, floor, ceil, trunc, sqrt

import pytest
from pytest import approx

from ..numeric import Probability


def test_probability_init():
    probability = Probability(0.2)

    assert probability == 0.2
    assert probability.log_value == log(0.2)

    Probability(log(0.2), log_value=True)

    assert probability == 0.2
    assert probability.log_value == log(0.2)

    with pytest.raises(ValueError, match="value must be greater than 0"):
        Probability(-0.2)


def test_probability_string():
    probability1 = Probability(0.2)
    probability2 = Probability(-1000, log_value=True)
    assert "Probability(0.2)" == repr(probability1)
    assert "Probability(-1000, log_value=True)" == repr(probability2)

    assert "0.2" == str(probability1)
    assert "exp(-1000)" == str(probability2)


def test_probability_comparison():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert probability1 == probability1
    assert not probability1 == -1
    assert probability1 != probability2
    assert probability1 != -1

    assert probability1 < probability2
    assert not probability1 < -1
    assert probability1 <= probability1
    assert probability1 <= probability2
    assert not probability1 <= -1

    assert probability2 > probability1
    assert probability2 > -1
    assert probability2 >= probability1
    assert probability2 >= probability2
    assert probability2 >= -1


def test_probability_addition():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert approx(0.5) == probability1 + probability2
    assert approx(0.5) == probability2 + probability1
    assert approx(0.5) == probability2 + 0.2
    assert approx(0.5) == 0.3 + probability1
    assert approx(0.4) == probability1 + probability1
    assert approx(0.1) == -0.2 + probability2
    assert approx(0.1) == probability2 + -0.2
    assert approx(0.2) == probability1 + 0
    assert approx(0.2) == 0 + probability1


def test_probability_subtraction():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert approx(0.1) == probability2 - probability1
    assert approx(0.1) == probability2 - 0.2
    assert approx(0.1) == 0.3 - probability1

    assert approx(0) == probability1 - probability1
    assert approx(0) == 0.3 - probability2
    assert approx(0.2) == probability1 - 0
    assert approx(0.5) == probability2 - -0.2

    assert approx(-0.3) == probability1 - 0.5
    assert approx(-0.1) == 0.2 - probability2
    assert approx(-0.2) == 0 - probability1
    assert approx(-0.5) == -0.3 - probability1

    assert approx(0.2) == 0.2 - Probability(0)


def test_probability_multiply():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert approx(0.06) == probability1 * probability2
    assert (probability1 * probability2).log_value == log(0.2) + log(0.3)

    assert approx(0.4) == probability1 * 2
    assert approx(0.6) == 2 * probability2

    assert approx(-0.4) == probability1 * -2
    assert approx(-0.6) == -2 * probability2


def test_probability_divide():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert approx(2/3) == probability1 / probability2
    assert (probability1 / probability2).log_value == log(0.2) - log(0.3)

    assert approx(2/3) == probability1 / 0.3
    assert approx(2/3) == 0.2 / probability2

    assert approx(-2/3) == probability1 / -0.3
    assert approx(-2/3) == -0.2 / probability2

    assert approx(0) == probability1 // probability2
    assert approx(0) == 0.2 // probability2
    assert approx(0) == probability1 // 0.3
    assert approx(-1) == -0.2 // probability2
    assert approx(-1) == probability1 // -0.3


def test_probability_mod():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert approx(0.1) == probability2 % probability1
    assert approx(0.1) == probability2 % 0.2
    assert approx(0.1) == 0.3 % probability1

    assert approx(-0.1) == probability2 % -0.2
    assert approx(0.1) == -0.3 % probability1


def test_probability_power():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert approx(0.2**0.3) == probability1 ** probability2
    assert approx(0.2**0.3) == 0.2 ** probability2
    assert approx(0.2**0.3) == probability1 ** 0.3

    assert approx(-0.2**0.3) == -0.2 ** probability2
    assert approx(0.2**-0.3) == probability1 ** -0.3


def test_probability_integral():
    probability = Probability(0.55)

    assert 0 == floor(probability)
    assert 0 == trunc(probability)
    assert 1 == ceil(probability)
    assert 1 == round(probability)
    assert approx(0.6) == round(probability, 1)


def test_probability_sign():
    probability = Probability(0.5)

    assert approx(0.5) == abs(probability)
    assert approx(0.5) == +probability
    assert approx(-0.5) == -probability


def test_probability_sum():
    probability1 = Probability(0.2)
    probability2 = Probability(0.3)

    assert approx(0.5) == Probability.sum((probability1, probability2))
    assert approx(0.5) == Probability.sum((0.2, probability2))
    assert approx(0.5) == Probability.sum((probability1, 0.3))
    assert approx(0.5) == Probability.sum((probability1, probability2, 0))

    assert approx(0.2) == Probability.sum((Probability(0), probability1))
    assert 0 == Probability.sum((Probability(0), ))

    assert 0 == Probability.sum([])


def test_probability_numpy_methods():
    probability = Probability(0.2)

    assert approx(sqrt(0.2)) == probability.sqrt()
    assert approx(log(0.2)) == probability.log()


def test_probability_hash():
    probability = Probability(0.2)

    assert hash(probability) == hash(0.2)

    probability = Probability(1E-100)**10
    # Float value zero, but should use custom hash method now
    assert hash(probability) != hash(0)
    # But equally, shouldn't match raw log value either.
    assert hash(probability) != hash(log(1E-100)*10)

    # Actual zero should match
    assert hash(Probability(0)) == hash(0)



# -*- coding: utf-8 -*-
from ..base import Property
from .state import State, StateMutableSequence


class GroundTruthState(State):
    """Ground Truth State type"""


class GroundTruthPath(StateMutableSequence):
    """Ground Truth Path type

    A :class:`~.StateMutableSequence` representing a track.
    """

    states = Property(
        [GroundTruthState],
        default=None,
        doc="List of groundtruth states to initialise path with. Default "
            "`None` which initialises with an empty list.")



from scipy.stats import multivariate_normal as mn

from .base import Hypothesiser
from ..base import Property
from ..types.detection import MissedDetection
from ..types.hypothesis import SingleProbabilityHypothesis
from ..types.multihypothesis import MultipleHypothesis
from ..types.numeric import Probability
from ..predictor import Predictor
from ..updater import Updater


class PDAHypothesiser(Hypothesiser):
    """Hypothesiser based on Probabilistic Data Association (PDA)

    Generate track predictions at detection times and calculate probabilities
    for all prediction-detection pairs for single prediction and multiple
    detections.
    """

    predictor = Property(
        Predictor,
        doc="Predict tracks to detection times")
    updater = Property(
        Updater,
        doc="Updater used to get measurement prediction")
    clutter_spatial_density = Property(
        float,
        doc="Spatial density of clutter - tied to probability of false "
            "detection")
    prob_detect = Property(
        Probability,
        default=Probability(0.85),
        doc="Target Detection Probability")
    prob_gate = Property(
        Probability,
        default=Probability(0.95),
        doc="Gate Probability - prob. gate contains true measurement "
            "if detected")

    def hypothesise(self, track, detections, timestamp):
        r"""Evaluate and return all track association hypotheses.

        For a given track and a set of N detections, return a
        MultipleHypothesis with N+1 detections (first detection is
        a 'MissedDetection'), each with an associated probability.
        Probabilities are assumed to be exhaustive (sum to 1) and mutually
        exclusive (two detections cannot be the correct association at the
        same time).

        Detection 0: missed detection, none of the detections are associated
        with the track.
        Detection :math:`i, i \in {1...N}`: detection i is associated
        with the track.

        The probabilities for these detections are calculated as follow:

        .. math::

          \beta_i(k) = \begin{cases}
                \frac{\mathcal{L}_{i}(k)}{1-P_{D}P_{G}+\sum_{j=1}^{m(k)}
                  \mathcal{L}_{j}(k)}, \quad i=1,...,m(k) \\
                \frac{1-P_{D}P_{G}}{1-P_{D}P_{G}+\sum_{j=1}^{m(k)}
                  \mathcal{L}_{j}(k)}, \quad i=0
                \end{cases}

        where

        .. math::

          \mathcal{L}_{i}(k) = \frac{\mathcal{N}[z_{i}(k);\hat{z}(k|k-1),
          S(k)]P_{D}}{\lambda}

        :math:`\lambda` is the clutter density

        :math:`P_{D}` is the detection probability

        :math:`P_{G}` is the gate probability

        :math:`\mathcal{N}[z_{i}(k);\hat{z}(k|k-1),S(k)]` is the likelihood
        ratio of the measurement :math:`z_{i}(k)` originating from the track
        target rather than the clutter.

        NOTE: Since all probabilities have the same denominator and are
        normalized later, the denominator can be discarded.

        References:

        [1] "The Probabilistic Data Association Filter: Estimation in the
        Presence of Measurement Origin Uncertainty" -
        https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=5338565

        [2] "Robotics 2 Data Association" (Lecture notes) -
        http://ais.informatik.uni-freiburg.de/teaching/ws10/robotics2/pdfs/rob2-15-dataassociation.pdf

        Parameters
        ----------
        track: :class:`~.Track`
            The track object to hypothesise on
        detections: :class:`list`
            A list of :class:`~Detection` objects, representing the available
            detections.
        timestamp: :class:`datetime.datetime`
            A timestamp used when evaluating the state and measurement
            predictions. Note that if a given detection has a non empty
            timestamp, then prediction will be performed according to
            the timestamp of the detection.

        Returns
        -------
        : :class:`~.MultipleHypothesis`
            A container of :class:`~SingleProbabilityHypothesis` objects

        """

        hypotheses = list()

        # Common state & measurement prediction
        prediction = self.predictor.predict(track.state, timestamp=timestamp)
        measurement_prediction = self.updater.get_measurement_prediction(
            prediction)

        # Missed detection hypothesis
        probability = Probability(1 - self.prob_detect*self.prob_gate)
        hypotheses.append(
            SingleProbabilityHypothesis(
                prediction,
                MissedDetection(timestamp=timestamp),
                probability,
                measurement_prediction))

        # True detection hypotheses
        for detection in detections:

            # Re-evaluate prediction
            prediction = self.predictor.predict(
                track.state, timestamp=detection.timestamp)

            # Compute measurement prediction and probability measure
            measurement_prediction = self.updater.get_measurement_prediction(
                prediction, detection.measurement_model)
            log_pdf = mn.logpdf(detection.state_vector.ravel(),
                                measurement_prediction.state_vector.ravel(),
                                measurement_prediction.covar)
            pdf = Probability(log_pdf, log_value=True)
            probability = (pdf * self.prob_detect)/self.clutter_spatial_density

            # True detection hypothesis
            hypotheses.append(
                SingleProbabilityHypothesis(
                    prediction,
                    detection,
                    probability,
                    measurement_prediction))

        return MultipleHypothesis(hypotheses, normalise=True, total_weight=1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

from stonesoup import __version__ as version

with open('README.md') as f:
    long_description = f.read()

setup(name='stonesoup',
      version=version,
      maintainer='Defence Science and Technology Laboratory UK',
      maintainer_email='oss@dstl.gov.uk',
      url='https://github.com/dstl/Stone-Soup',
      description='A tracking and state estimation framework',
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3 :: Only',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering',
      ],
      packages=find_packages(exclude=('docs', '*.tests')),
      install_requires=[
          'ruamel.yaml>=0.15.45', 'scipy', 'matplotlib', 'utm', 'pymap3d'],
      extras_require={
          'dev': [
              'pytest-flake8', 'pytest-cov', 'Sphinx', 'sphinx_rtd_theme',
              'setuptools>=30'],
      },
      )



# coding: utf-8
import datetime

import numpy as np
import pytest

from ...types.state import State
from ...platform.simple import SensorPlatform
from ...models.transition.linear import (
    ConstantVelocity, CombinedLinearGaussianTransitionModel)
from ...sensor.radar import RadarRangeBearing
from ...types.array import StateVector, CovarianceMatrix

# Input arguments
# TODO: pytest parametarization


def get_3d_expected(i):
    if i == 0:
        # static platform or X velocity
        return np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [-1, 0, 0],
                         [0, -1, 0], [0, 0, 1], [0, 0, -1]])
    elif i == 1:
        # y-axis motion
        return np.array([[0, 0, 0], [0, 1, 0], [-1, 0, 0], [0, -1, 0],
                         [1, 0, 0], [0, 0, 1], [0, 0, -1]])
    elif i == 2:
        # negative x-axis motion
        return np.array([[0, 0, 0], [-1, 0, 0], [0, -1, 0], [1, 0, 0],
                         [0, 1, 0], [0, 0, 1], [0, 0, -1]])
    elif i == 3:
        # negative y-axis motion
        return np.array([[0, 0, 0], [0, -1, 0], [1, 0, 0], [0, 1, 0],
                         [-1, 0, 0], [0, 0, 1], [0, 0, -1]])
    elif i == 4:
        # x-y motion
        return np.array([[0, 0, 0], [1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [1/np.sqrt(2), -1/np.sqrt(2), 0], [0, 0, 1],
                         [0, 0, -1]])
    elif i == 5:
        # neg x- neg y motion
        return np.array([[0, 0, 0], [-1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-1/np.sqrt(2), 1/np.sqrt(2), 0], [0, 0, 1],
                         [0, 0, -1]])
    elif i == 6:
        # pos x- neg y motion
        return np.array([[0, 0, 0], [1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-1/np.sqrt(2), -1/np.sqrt(2), 0], [0, 0, 1],
                         [0, 0, -1]])
    elif i == 7:
        # neg x- pos y motion
        return np.array([[0, 0, 0], [-1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [1/np.sqrt(2), 1/np.sqrt(2), 0], [0, 0, 1],
                         [0, 0, -1]])
    elif i == 8:
        # "z vel"
        return np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 0, -1],
                         [0, -1, 0], [-1, 0, 0], [1, 0, 0]])
    elif i == 9:
        # "-z vel"
        return np.array([[0, 0, 0], [0, 0, -1], [0, 1, 0], [0, 0, 1],
                         [0, -1, 0], [1, 0, 0], [-1, 0, 0]])
    elif i == 10:
        # "y.z vel"
        return np.array([[0, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [-1, 0, 0], [0, -1/np.sqrt(2), -1/np.sqrt(2)],
                         [1, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, 1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 11:
        # "y.-z vel"
        return np.array([[0, 0, 0], [0,  1/np.sqrt(2), -1/np.sqrt(2)],
                         [-1, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [1, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, -1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 12:
        # "-y.z vel"
        return np.array([[0, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [1, 0, 0], [0, 1/np.sqrt(2), -1/np.sqrt(2)],
                         [-1, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, -1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 13:
        # "-y.-z vel"
        return np.array([[0, 0, 0], [0, -1/np.sqrt(2), -1/np.sqrt(2)],
                         [1, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [-1, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, 1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 14:
        # x.z vel
        return np.array([[0, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [-1, 0, 0], [0, -1/np.sqrt(2), -1/np.sqrt(2)],
                         [1, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, 1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 15:
        # -x.z vel
        return np.array([[0, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [1, 0, 0], [0, 1/np.sqrt(2), -1/np.sqrt(2)],
                         [-1, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, -1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 16:
        # x.-z vel
        return np.array([[0, 0, 0], [0, 1/np.sqrt(2), -1/np.sqrt(2)],
                         [-1, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [1, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, -1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 17:
        # -x,-z vel
        return np.array([[0, 0, 0], [0, -1/np.sqrt(2), -1/np.sqrt(2)],
                         [1, 0, 0], [0, 1/np.sqrt(2), 1/np.sqrt(2)],
                         [-1, 0, 0], [0, -1/np.sqrt(2), 1/np.sqrt(2)],
                         [0, 1/np.sqrt(2), -1/np.sqrt(2)]])
    elif i == 18:
        # x.y.z vel
        a = np.cos(np.arctan2(1, np.sqrt(2)) * -1)
        b = np.sin(np.arctan2(1, np.sqrt(2)) * -1) / np.sqrt(2)
        return np.array([[0, 0, 0], [1/np.sqrt(3), 1/np.sqrt(3), 1/np.sqrt(3)],
                         [-1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-1/np.sqrt(3), -1/np.sqrt(3), -1/np.sqrt(3)],
                         [1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [b, b, a], [-b, -b, -a]])
    elif i == 19:
        # -x.-y.-z vel
        a = np.cos(np.arctan2(-1, np.sqrt(2)) * -1)
        b = np.sin(np.arctan2(-1, np.sqrt(2)) * -1) / np.sqrt(2)
        return np.array([[0, 0, 0],
                         [-1/np.sqrt(3), -1/np.sqrt(3), -1/np.sqrt(3)],
                         [1/np.sqrt(2), -1/np.sqrt(2), 0],
                         [1/np.sqrt(3), 1/np.sqrt(3), 1/np.sqrt(3)],
                         [-1/np.sqrt(2), 1/np.sqrt(2), 0],
                         [-b, -b, a], [b, b, -a]])


@pytest.fixture
def radars_2d():
    # Generate 5 radar models for testing purposes
    noise_covar = CovarianceMatrix(np.array([[0.015, 0],
                                             [0, 0.1]]))

    # define arbitrary sensor origin
    radar1_position = StateVector(np.array(([[100], [100]])))
    radar1_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar2_position = StateVector(np.array(([[100], [100]])))
    radar2_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar3_position = StateVector(np.array(([[100], [100]])))
    radar3_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar4_position = StateVector(np.array(([[100], [100]])))
    radar4_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar5_position = StateVector(np.array(([[100], [100]])))
    radar5_orientation = StateVector(np.array(([[0], [0], [0]])))

    measurement_mapping = np.array([0, 2])

    # Create 5 simple radar sensor objects
    radar1 = RadarRangeBearing(
        position=radar1_position,
        orientation=radar1_orientation,
        ndim_state=4,
        mapping=measurement_mapping,
        noise_covar=noise_covar,
    )

    radar2 = RadarRangeBearing(
        position=radar2_position,
        orientation=radar2_orientation,
        ndim_state=4,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )
    radar3 = RadarRangeBearing(
        position=radar3_position,
        orientation=radar3_orientation,
        ndim_state=4,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )

    radar4 = RadarRangeBearing(
        position=radar4_position,
        orientation=radar4_orientation,
        ndim_state=4,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )
    radar5 = RadarRangeBearing(
        position=radar5_position,
        orientation=radar5_orientation,
        ndim_state=4,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )

    return [radar1, radar2, radar3, radar4, radar5]


@pytest.fixture
def radars_3d():
    # Generate 7 radar models for testing purposes
    noise_covar = CovarianceMatrix(np.array([[0.015, 0],
                                             [0, 0.1]]))

    # Note 1 - the radar position is irrelevant once mounted
    radar1_position = StateVector(np.array(([[100], [100], [100]])))
    radar1_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar2_position = StateVector(np.array(([[100], [100], [100]])))
    radar2_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar3_position = StateVector(np.array(([[100], [100], [100]])))
    radar3_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar4_position = StateVector(np.array(([[100], [100], [100]])))
    radar4_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar5_position = StateVector(np.array(([[100], [100], [100]])))
    radar5_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar6_position = StateVector(np.array(([[100], [100], [100]])))
    radar6_orientation = StateVector(np.array(([[0], [0], [0]])))
    radar7_position = StateVector(np.array(([[100], [100], [100]])))
    radar7_orientation = StateVector(np.array(([[0], [0], [0]])))

    measurement_mapping = np.array([0, 2, 4])

    # Create 5 simple radar sensor objects
    radar1 = RadarRangeBearing(
        position=radar1_position,
        orientation=radar1_orientation,
        ndim_state=6,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )

    radar2 = RadarRangeBearing(
        position=radar2_position,
        orientation=radar2_orientation,
        ndim_state=6,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )
    radar3 = RadarRangeBearing(
        position=radar3_position,
        orientation=radar3_orientation,
        ndim_state=6,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )

    radar4 = RadarRangeBearing(
        position=radar4_position,
        orientation=radar4_orientation,
        ndim_state=6,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )
    radar5 = RadarRangeBearing(
        position=radar5_position,
        orientation=radar5_orientation,
        ndim_state=6,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )
    radar6 = RadarRangeBearing(
        position=radar6_position,
        orientation=radar6_orientation,
        ndim_state=6,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )
    radar7 = RadarRangeBearing(
        position=radar7_position,
        orientation=radar7_orientation,
        ndim_state=6,
        mapping=measurement_mapping,
        noise_covar=noise_covar
    )
    return [radar1, radar2, radar3, radar4, radar5, radar6, radar7]


@pytest.fixture(scope='session')
def mounting_offsets_2d():
    # Generate sensor mounting offsets for testing purposes
    return np.array([[0, 0],
                     [1, 0],
                     [0, 1],
                     [-1, 0],
                     [0, -1]])


@pytest.fixture(scope='session')
def mounting_offsets_3d():
    # Generate sensor mounting offsets for testing purposes
    return np.array([[0, 0, 0],
                     [1, 0, 0],
                     [0, 1, 0],
                     [-1, 0, 0],
                     [0, -1, 0],
                     [0, 0, 1],
                     [0, 0, -1]])


@pytest.fixture(params=[True, False], ids=["Moving", "Static"])
def move(request):
    return request.param


testdata_2d = [
    np.array([[0], [0], [0], [0]]),
    np.array([[10], [0], [0], [0]]),
    np.array([[0], [1], [0], [0]]),
    np.array([[0], [0], [0], [1]]),
    np.array([[0], [-1], [0], [0]]),
    np.array([[0], [0], [0], [-1]]),
    np.array([[0], [1], [0], [1]]),
    np.array([[0], [-1], [0], [-1]]),
    np.array([[0], [1], [0], [-1]]),
    np.array([[0], [-1], [0], [1]])
]

expected_2d = [
    # static platform or X velocity
    np.array([[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]]),
    # static platform or X velocity
    np.array([[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]]),
    # static platform or X velocity
    np.array([[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]]),
    # y-axis motion
    np.array([[0, 0], [0, 1], [-1, 0], [0, -1], [1, 0]]),
    # negative x-axis motion
    np.array([[0, 0], [-1, 0], [0, -1], [1, 0], [0, 1]]),
    # negative y-axis motion
    np.array([[0, 0], [0, -1], [1, 0], [0, 1], [-1, 0]]),
    # x-y motion
    np.array([[0, 0], [1/np.sqrt(2), 1/np.sqrt(2)],
              [-1/np.sqrt(2), 1/np.sqrt(2)], [-1/np.sqrt(2), -1/np.sqrt(2)],
              [1/np.sqrt(2), -1/np.sqrt(2)]]),
    # neg x- neg y motion
    np.array([[0, 0], [-1/np.sqrt(2), -1/np.sqrt(2)],
              [1/np.sqrt(2), -1/np.sqrt(2)], [1/np.sqrt(2), 1/np.sqrt(2)],
              [-1/np.sqrt(2), 1/np.sqrt(2)]]),
    # pos x- neg y motion
    np.array([[0, 0], [1/np.sqrt(2), -1/np.sqrt(2)],
              [1/np.sqrt(2), 1/np.sqrt(2)], [-1/np.sqrt(2), 1/np.sqrt(2)],
              [-1/np.sqrt(2), -1/np.sqrt(2)]]),
    # neg x- pos y motion
    np.array([[0, 0], [-1/np.sqrt(2), 1/np.sqrt(2)],
              [-1/np.sqrt(2), -1/np.sqrt(2)], [1/np.sqrt(2), -1/np.sqrt(2)],
              [1/np.sqrt(2), 1/np.sqrt(2)]])
]


@pytest.mark.parametrize(
    'state, expected', zip(testdata_2d, expected_2d),
    ids=["Static", "pos offset", "x vel", "y vel", "-x vel", "-y vel",
         "x,y vel", "-x,-y vel", "x,-y vel", "-x,y vel"])
def test_2d_platform(state, expected, move, radars_2d, mounting_offsets_2d):
    # Define time related variables
    timestamp = datetime.datetime.now()
    # Define transition model and position for platform
    model_1d = ConstantVelocity(0.0)  # zero noise so pure movement
    trans_model = CombinedLinearGaussianTransitionModel(
        [model_1d] * (radars_2d[0].ndim_state // 2))
    platform_state = State(state, timestamp)

    # This defines the mapping to the platforms state vector (i.e. x and y)
    mounting_mappings = np.array([[0, 2]])
    # create a platform with the simple radar mounted
    platform = SensorPlatform(
        state=platform_state,
        transition_model=trans_model,
        sensors=radars_2d,
        mounting_offsets=mounting_offsets_2d,
        mounting_mappings=mounting_mappings
    )
    if move:
        # Move the platform
        platform.move(timestamp + datetime.timedelta(seconds=2))
    sensor_positions_test(expected, platform)


testdata_3d = [
    (np.array([[0], [0], [0], [0], [0], [0]]), get_3d_expected(0)),
    (np.array([[10], [0], [0], [0], [0], [0]]), get_3d_expected(0)),
    (np.array([[0], [1], [0], [0], [0], [0]]), get_3d_expected(0)),
    (np.array([[0], [0], [0], [1], [0], [0]]), get_3d_expected(1)),
    (np.array([[0], [-1], [0], [0], [0], [0]]), get_3d_expected(2)),
    (np.array([[0], [0], [0], [-1], [0], [0]]), get_3d_expected(3)),
    (np.array([[0], [1], [0], [1], [0], [0]]), get_3d_expected(4)),
    (np.array([[0], [-1], [0], [-1], [0], [0]]), get_3d_expected(5)),
    (np.array([[0], [1], [0], [-1], [0], [0]]), get_3d_expected(6)),
    (np.array([[0], [-1], [0], [1], [0], [0]]), get_3d_expected(7)),
    (np.array([[0], [0], [0], [0], [0], [1]]), get_3d_expected(8)),
    (np.array([[0], [0], [0], [0], [0], [-1]]), get_3d_expected(9)),
    (np.array([[0], [0], [0], [1], [0], [1]]), get_3d_expected(10)),
    (np.array([[0], [0], [0], [1], [0], [-1]]), get_3d_expected(11)),
    (np.array([[0], [0], [0], [-1], [0], [1]]), get_3d_expected(12)),
    (np.array([[0], [0], [0], [-1], [0], [-1]]), get_3d_expected(13)),
    (np.array([[0], [1], [0], [0], [0], [1]]), get_3d_expected(14)),
    (np.array([[0], [-1], [0], [0], [0], [1]]), get_3d_expected(15)),
    (np.array([[0], [1], [0], [0], [0], [-1]]), get_3d_expected(16)),
    (np.array([[0], [-1], [0], [0], [0], [-1]]), get_3d_expected(17)),
    (np.array([[0], [1], [0], [1], [0], [1]]), get_3d_expected(18)),
    (np.array([[0], [-1], [0], [-1], [0], [-1]]), get_3d_expected(19))
]


@pytest.mark.parametrize('state, expected', testdata_3d, ids=[
    "Static", "pos offset", "x vel", "y vel", "-x vel", "-y vel", "x,y vel",
    "-x,-y vel", "x,-y vel", "-x,y vel", "z vel", "-z vel", "y.z vel",
    "y.-z vel", "-y.z vel", "-y.-z vel", "x.z vel", "-x.z vel", "x.-z vel",
    "-x,-z vel", "x,y,z vel", "-x,-y,-z vel"
])
def test_3d_platform(state, expected, move, radars_3d, mounting_offsets_3d):
    # Define time related variables
    timestamp = datetime.datetime.now()
    # Define transition model and position for platform
    model_1d = ConstantVelocity(0.0)  # zero noise so pure movement
    trans_model = CombinedLinearGaussianTransitionModel(
        [model_1d] * (radars_3d[0].ndim_state // 2))
    platform_state = State(state, timestamp)

    # This defines the mapping to the platforms state vector (i.e. x and y)
    mounting_mappings = np.array([[0, 2, 4]])
    # create a platform with the simple radar mounted
    platform = SensorPlatform(
        state=platform_state,
        transition_model=trans_model,
        sensors=radars_3d,
        mounting_offsets=mounting_offsets_3d,
        mounting_mappings=mounting_mappings
    )
    if move:
        # Move the platform
        platform.move(timestamp + datetime.timedelta(seconds=2))
    sensor_positions_test(expected, platform)


def sensor_positions_test(expected, platform):
    """
    This function asserts that the sensor positions on the platform have been
    correctly updated when the platform has been moved or sensor mounted on the
    platform.

    :param expected: nD array of expected sensor position post rotation
    :param platform: platform object
    :return:
    """
    expected_radar_position = np.zeros(
        [len(platform.sensors), platform.mounting_offsets.shape[1]])
    for i in range(len(platform.sensors)):
        radar_position = platform.sensors[i].position
        for j in range(platform.mounting_offsets.shape[1]):
            expected_radar_position[i, j] = (expected[i, j] +
                                             platform.state.state_vector[
                                                 platform.mounting_mappings[
                                                     i, j]])
        assert (np.allclose(expected_radar_position[i, j], radar_position[j]))



# -*- coding: utf-8 -*-

import datetime

import numpy as np
import pytest

from ..association import Association, AssociationPair, AssociationSet, \
    SingleTimeAssociation, TimeRangeAssociation
from ..detection import Detection
from ..time import TimeRange


def test_association():
    with pytest.raises(TypeError):
        Association()

    objects = {Detection(np.array([[1], [2]])),
               Detection(np.array([[3], [4]])),
               Detection(np.array([[5], [6]]))}

    assoc = Association(objects)

    assert assoc.objects == objects


def test_associationpair():
    with pytest.raises(TypeError):
        AssociationPair()

    objects = [Detection(np.array([[1], [2]])),
               Detection(np.array([[3], [4]])),
               Detection(np.array([[5], [6]]))]

    # Over 3 objects
    with pytest.raises(ValueError):
        AssociationPair(set(objects))

    # Under 2 objects
    with pytest.raises(ValueError):
        AssociationPair({objects[0]})

    # 2 objects
    assoc = AssociationPair(set(objects[:2]))
    np.array_equal(assoc.objects, set(objects[:2]))


def test_singletimeassociation():
    with pytest.raises(TypeError):
        SingleTimeAssociation()

    objects = {Detection(np.array([[1], [2]])),
               Detection(np.array([[3], [4]])),
               Detection(np.array([[5], [6]]))}
    timestamp = datetime.datetime(2018, 3, 1, 5, 3, 35)

    assoc = SingleTimeAssociation(objects=objects, timestamp=timestamp)

    assert assoc.objects == objects
    assert assoc.timestamp == timestamp


def test_timerangeassociation():
    with pytest.raises(TypeError):
        TimeRangeAssociation()

    objects = {Detection(np.array([[1], [2]])),
               Detection(np.array([[3], [4]])),
               Detection(np.array([[5], [6]]))}
    timestamp1 = datetime.datetime(2018, 3, 1, 5, 3, 35)
    timestamp2 = datetime.datetime(2018, 3, 1, 5, 8, 35)
    timerange = TimeRange(start_timestamp=timestamp1, end_timestamp=timestamp2)

    assoc = TimeRangeAssociation(objects=objects, time_range=timerange)

    assert assoc.objects == objects
    assert assoc.time_range == timerange


def test_associationset():
    # Set up some dummy variables
    timestamp1 = datetime.datetime(2018, 3, 1, 5, 3, 35)
    timestamp2 = datetime.datetime(2018, 3, 1, 5, 8, 35)
    timerange = TimeRange(start_timestamp=timestamp1, end_timestamp=timestamp2)

    objects_list = [Detection(np.array([[1], [2]])),
                    Detection(np.array([[3], [4]])),
                    Detection(np.array([[5], [6]]))]

    assoc1 = SingleTimeAssociation(objects=set(objects_list),
                                   timestamp=timestamp1)

    assoc2 = TimeRangeAssociation(objects=set(objects_list[1:]),
                                  time_range=timerange)

    assoc_set = AssociationSet({assoc1, assoc2})

    assert assoc_set.associations == {assoc1, assoc2}

    # Test associations including objects

    # Object only present in object 1
    assert assoc_set.associations_including_objects(objects_list[0]) \
        == {assoc1}
    # Object present in both
    assert assoc_set.associations_including_objects(objects_list[1]) \
        == {assoc1, assoc2}
    # Object present in neither
    assert not assoc_set.associations_including_objects(
        Detection(np.array([[0], [0]])))

    # Test associations including timestamp

    # Timestamp present in one object
    assert assoc_set.associations_at_timestamp(timestamp2) \
        == {assoc2}
    # Timestamp present in both
    assert assoc_set.associations_at_timestamp(timestamp1) \
        == {assoc1, assoc2}
    # Timestamp not present in either
    timestamp3 = datetime.datetime(2018, 3, 1, 6, 8, 35)
    assert not assoc_set.associations_at_timestamp(timestamp3)



# -*- coding: utf-8 -*-
import datetime

import pytest
import numpy as np

from ...types.state import State
from ..simple import SimpleDetectionSimulator, SingleTargetGroundTruthSimulator


@pytest.fixture(params=[datetime.timedelta(seconds=1),
                        datetime.timedelta(seconds=10),
                        datetime.timedelta(minutes=1)])
def timestep(request):
    return request.param


def test_simple_detection_simulator(
        transition_model, measurement_model, timestep):
    initial_state = State(
        np.array([[0], [0], [0], [0]]), timestamp=datetime.datetime.now())
    groundtruth = SingleTargetGroundTruthSimulator(
        transition_model, initial_state, timestep)
    meas_range = np.array([[-1, 1], [-1, 1]]) * 5000
    detector = SimpleDetectionSimulator(
        groundtruth, measurement_model, meas_range, clutter_rate=3)

    total_detections = set()
    clutter_detections = set()
    for step, (time, detections) in enumerate(detector.detections_gen()):
        total_detections |= detections
        clutter_detections |= detector.clutter_detections

        # Check time increments correctly
        assert time == initial_state.timestamp + step * timestep

    # Check both real and clutter detections are generated
    assert len(total_detections) > len(clutter_detections)

    # Check clutter is generated within specified bounds
    for clutter in clutter_detections:
        assert (meas_range[:, 0] <= clutter.state_vector.ravel()).all()
        assert (meas_range[:, 1] >= clutter.state_vector.ravel()).all()

    assert detector.clutter_spatial_density == 4.8e-15



# -*- coding: utf-8 -*-
import datetime

import pytest

from ..tracktotrack import EuclideanTrackToTrack, EuclideanTrackToTruth
from ...models.measurement.linear import LinearGaussian
from ...types.state import State
from ...types.track import Track


@pytest.fixture
def tracks():
    start_time = datetime.datetime(2019, 1, 1, 14, 0, 0)

    tracks = [Track(states=[
        State(state_vector=[[i], [i]],
              timestamp=start_time + datetime.timedelta(seconds=i))
        for i in range(10)])]

    # 2nd track should be associated with track1 from the second timestamp to
    # the 6th
    tracks.append(Track(states=[
        State(state_vector=[[20], [20]],
              timestamp=start_time)]
        + [State(state_vector=[[i + 2], [i + 2]],
                 timestamp=start_time + datetime.timedelta(seconds=i))
           for i in range(1, 7)]))

    # 3rd is at a different time so shoult not associate with anything
    tracks.append(Track(states=[
        State(state_vector=[[i], [i]],
              timestamp=start_time + datetime.timedelta(seconds=i + 20))
        for i in range(10)]))

    # 4th is outside the association threshold
    tracks.append(Track(states=[
        State(state_vector=[[i + 20], [i + 20]],
              timestamp=start_time + datetime.timedelta(seconds=i))
        for i in range(10)]))

    return tracks


def test_euclidiantracktotrack(tracks):
    associator = EuclideanTrackToTrack(
        association_threshold=10,
        consec_pairs_confirm=3,
        consec_misses_end=2,
        measurement_model_track1=LinearGaussian(2, [0, 1], None),
        measurement_model_track2=LinearGaussian(2, [0, 1], None))
    start_time = datetime.datetime(2019, 1, 1, 14, 0, 0)

    association_set = associator.associate_tracks({tracks[0], tracks[2]},
                                                  {tracks[1], tracks[3]})

    assert len(association_set.associations) == 1
    assoc = list(association_set.associations)[0]
    assert set(assoc.objects) == {tracks[0], tracks[1]}
    assert assoc.time_range.start_timestamp \
        == start_time + datetime.timedelta(seconds=1)
    assert assoc.time_range.end_timestamp \
        == start_time + datetime.timedelta(seconds=6)


def test_euclidiantracktotruth(tracks):
    associator = EuclideanTrackToTruth(
        association_threshold=10,
        consec_pairs_confirm=3,
        consec_misses_end=2,
        measurement_model_track=LinearGaussian(2, [0, 1], None),
        measurement_model_truth=LinearGaussian(2, [0, 1], None))
    start_time = datetime.datetime(2019, 1, 1, 14, 0, 0)

    association_set = associator.associate_tracks(
        truth_set={tracks[0]}, tracks_set={tracks[2], tracks[1], tracks[3]})

    assert len(association_set.associations) == 1
    assoc = list(association_set.associations)[0]
    assert set(assoc.objects) == {tracks[0], tracks[1]}
    assert assoc.time_range.start_timestamp == start_time + datetime.timedelta(
        seconds=1)
    assert assoc.time_range.end_timestamp == start_time + datetime.timedelta(
        seconds=6)



# coding: utf-8
import datetime

import scipy as sp
from scipy.stats import multivariate_normal

from ..linear import ConstantTurn


def test_ctmodel():
    """ ConstantTurn Transition Model test """
    state_vec = sp.array([[3.0], [1.0], [2.0], [1.0]])
    noise_diff_coeffs = sp.array([0.01, 0.01])
    turn_rate = 0.1
    base(ConstantTurn, state_vec, noise_diff_coeffs, turn_rate)


def base(model, state_vec, noise_diff_coeffs, turn_rate):
    """ Base test for n-dimensional ConstantAcceleration Transition Models """

    # Create an ConstantTurn model object
    model = model
    model_obj = model(noise_diff_coeffs=noise_diff_coeffs, turn_rate=turn_rate)

    # State related variables
    state_vec = state_vec
    old_timestamp = datetime.datetime.now()
    timediff = 1  # 1sec
    new_timestamp = old_timestamp + datetime.timedelta(seconds=timediff)
    time_interval = new_timestamp - old_timestamp

    # Model-related components
    noise_diff_coeffs = noise_diff_coeffs  # m/s^3
    turn_rate = turn_rate
    turn_ratedt = turn_rate*timediff
    F = sp.array(
            [[1, sp.sin(turn_ratedt) / turn_rate,
              0, -(1 - sp.cos(turn_ratedt)) / turn_rate],
             [0, sp.cos(turn_ratedt),
              0, -sp.sin(turn_ratedt)],
             [0, (1 - sp.cos(turn_ratedt)) / turn_rate,
              1, sp.sin(turn_ratedt) / turn_rate],
             [0, sp.sin(turn_ratedt),
              0, sp.cos(turn_ratedt)]])

    qx = noise_diff_coeffs[0]
    qy = noise_diff_coeffs[1]
    Q = sp.array([[sp.power(qx, 2) * sp.power(timediff, 3) / 3,
                   sp.power(qx, 2) * sp.power(timediff, 2) / 2,
                   0,
                   0],
                  [sp.power(qx, 2) * sp.power(timediff, 2) / 2,
                   sp.power(qx, 2) * timediff,
                   0,
                   0],
                  [0,
                   0,
                   sp.power(qy, 2) * sp.power(timediff, 3) / 3,
                   sp.power(qy, 2) * sp.power(timediff, 2) / 2],
                  [0,
                   0,
                   sp.power(qy, 2) * sp.power(timediff, 2) / 2,
                   sp.power(qy, 2) * timediff]])

    # Ensure ```model_obj.transfer_function(time_interval)``` returns F
    assert sp.array_equal(F, model_obj.matrix(
        timestamp=new_timestamp, time_interval=time_interval))

    # Ensure ```model_obj.covar(time_interval)``` returns Q
    assert sp.array_equal(Q, model_obj.covar(
        timestamp=new_timestamp, time_interval=time_interval))

    # Propagate a state vector throught the model
    # (without noise)
    new_state_vec_wo_noise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=0)
    assert sp.array_equal(new_state_vec_wo_noise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (without noise)
    prob = model_obj.pdf(new_state_vec_wo_noise,
                         state_vec,
                         timestamp=new_timestamp,
                         time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_wo_noise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throughout the model
    # (with internal noise)
    new_state_vec_w_inoise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval)
    assert not sp.array_equal(new_state_vec_w_inoise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model_obj.pdf(new_state_vec_w_inoise,
                         state_vec,
                         timestamp=new_timestamp,
                         time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_inoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throught the model
    # (with external noise)
    noise = model_obj.rvs(timestamp=new_timestamp, time_interval=time_interval)
    new_state_vec_w_enoise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=noise)
    assert sp.array_equal(new_state_vec_w_enoise, F@state_vec+noise)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model_obj.pdf(new_state_vec_w_enoise, state_vec,
                         timestamp=new_timestamp, time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_enoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)



# -*- coding: utf-8 -*-
"""Base classes for different Readers."""
from abc import abstractmethod

from ..base import Base


class Reader(Base):
    """Reader base class"""


class DetectionReader(Reader):
    """Detection Reader base class"""

    @property
    @abstractmethod
    def detections(self):
        """The detections at the current time step.

        This is the set of detections last returned by the
        :meth:`detections_gen` generator, to allow other components, like
        metrics, to access the data.
        """
        raise NotImplementedError

    @abstractmethod
    def detections_gen(self):
        """Returns a generator of detections for each time step.

        Yields
        ------
        : :class:`datetime.datetime`
            Datetime of current time step
        : set of :class:`~.Detection`
            Detections generate in the time step
        """
        raise NotImplementedError


class GroundTruthReader(Reader):
    """Ground Truth Reader base class"""

    @property
    @abstractmethod
    def groundtruth_paths(self):
        """The groundtruth paths at the current time step.

        This is the set of groundtruth paths last returned by the
        :meth:`groundtruth_paths_gen` generator, to allow other components,
        like metrics, to access the data.
        """
        raise NotImplementedError

    @abstractmethod
    def groundtruth_paths_gen(self):
        """Returns a generator of ground truth paths for each time step.

        Yields
        ------
        : :class:`datetime.datetime`
            Datetime of current time step
        : set of :class:`~.GroundTruthPath`
            Ground truth paths existing in the time step
        """
        raise NotImplementedError


class SensorDataReader(Reader):
    """Sensor Data Reader base class"""

    @property
    @abstractmethod
    def sensor_data(self):
        """The sensor data at the current time step.

        This is the set of sensor data last returned by the
        :meth:`sensor_data_gen` generator, to allow other components, like
        metrics, to access the data.
        """
        raise NotImplementedError

    @abstractmethod
    def sensor_data_gen(self):
        """Returns a generator of sensor data for each time step.

        Yields
        ------
        : :class:`datetime.datetime`
            Datetime of current time step
        : set of :class:`~.SensorData`
            Sensor data generated in the time step
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
from math import log, log1p, exp, trunc, ceil, floor
from numbers import Real, Integral


class Probability(Real):
    """Probability class.

    Similar to a float, but value stored as natural log value internally.
    All operations are attempted with log values where possible, and failing
    that a float will be returned instead.
    """

    def __init__(self, value, *, log_value=False):
        if log_value:
            self._log_value = value
        else:
            try:
                self._log_value = self._log(value)
            except ValueError:
                raise ValueError("value must be greater than 0")

    @property
    def log_value(self):
        return self._log_value

    @staticmethod
    def _log(other):
        if isinstance(other, Probability):
            return other.log_value
        elif other == 0:
            return float("-inf")
        else:
            return log(other)

    def __hash__(self):
        value = float(self)
        if value == 0 and self.log_value != float("-inf"):  # Too close to zero
            # Add string so doesn't have same hash as the log value itself.
            return hash(('log', self._log_value))
        else:
            return hash(value)

    def __eq__(self, other):
        if other < 0:
            return False
        return self.log_value == self._log(other)

    def __le__(self, other):
        if other < 0:
            return False
        return self.log_value <= self._log(other)

    def __lt__(self, other):
        if other < 0:
            return False
        return self.log_value < self._log(other)

    def __ge__(self, other):
        if other < 0:
            return True
        return self.log_value >= self._log(other)

    def __gt__(self, other):
        if other < 0:
            return True
        return self.log_value > self._log(other)

    def __add__(self, other):
        if other < 0:
            return self - -other

        log_other = self._log(other)
        if self.log_value > log_other:
            log_l, log_s = self.log_value, log_other
        elif self.log_value < log_other:
            log_l, log_s = log_other, self.log_value
        else:  # Must be equal, so just double value
            return self * 2

        if log_s == float("-inf"):  # Just return largest value
            return Probability(log_l, log_value=True)

        return Probability(log_l + log1p(exp(log_s - log_l)),
                           log_value=True)

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        if other < 0:
            return self + -other

        log_other = self._log(other)
        if self.log_value > log_other:
            log_l, log_s = self.log_value, log_other
        elif self.log_value < log_other:  # Result will be negative
            return float(self) - other
        else:  # Must be equal, so return 0
            return Probability(float("-inf"), log_value=True)

        if log_s == float("-inf"):  # Just return largest value
            return Probability(log_l, log_value=True)

        exp_diff = exp(log_s - log_l)
        if exp_diff == 1:  # Diff too small, so result is effectively zero
            return Probability(float("-inf"), log_value=True)

        return Probability(log_l + log1p(-exp_diff),
                           log_value=True)

    def __rsub__(self, other):
        if other < 0:  # Result will be negative
            return other + -float(self)

        log_other = self._log(other)
        if log_other > self.log_value:
            log_l, log_s = log_other, self.log_value
        elif log_other < self.log_value:  # Result will be negative
            return other + -float(self)
        else:  # Must be equal, so return 0
            return Probability(float("-inf"), log_value=True)

        if log_s == float("-inf"):  # Just return largest value
            return Probability(log_l, log_value=True)

        exp_diff = exp(log_s - log_l)
        if exp_diff == 1:  # Diff too small, so result is effectively zero
            return Probability(float("-inf"), log_value=True)

        return Probability(log_l + log1p(-exp_diff),
                           log_value=True)

    def __mul__(self, other):
        try:
            return Probability(self.log_value + self._log(other),
                               log_value=True)
        except ValueError:
            return float(self) * other

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        try:
            return Probability(self.log_value - self._log(other),
                               log_value=True)
        except ValueError:
            return float(self) / other

    def __rtruediv__(self, other):
        try:
            return Probability(self._log(other) - self.log_value,
                               log_value=True)
        except ValueError:
            return other / float(self)

    def __floordiv__(self, other):
        return floor(self / other)

    def __rfloordiv__(self, other):
        return floor(other / self)

    def __mod__(self, other):
        try:
            return Probability(float(self) % other)
        except ValueError:
            return float(self) % other

    def __rmod__(self, other):
        return Probability(other % float(self))

    def __pow__(self, exponent):
        if isinstance(exponent, Probability):
            exponent = float(exponent)
        return Probability(exponent * self.log_value, log_value=True)

    def __rpow__(self, base):
        return Probability(base ** float(self))

    def __neg__(self):
        return -float(self)

    def __pos__(self):
        return Probability(self)

    def __abs__(self):
        return Probability(self)

    def __float__(self):
        return exp(self.log_value)

    def __round__(self, ndigits=None):
        value = round(float(self), ndigits)
        if isinstance(value, Integral):
            return value
        else:
            return Probability(value)

    def __trunc__(self):
        return trunc(float(self))

    def __floor__(self):
        return floor(float(self))

    def __ceil__(self):
        return ceil(float(self))

    def __repr__(self):
        value = float(self)
        if value == 0 and self.log_value != float("-inf"):  # Too close to zero
            return "Probability({!r}, log_value=True)".format(self.log_value)
        else:
            return "Probability({!r})".format(float(self))

    def __str__(self):
        value = float(self)
        if value == 0 and self.log_value != float("-inf"):  # Too close to zero
            return "exp({})".format(self.log_value)
        else:
            return str(value)

    def sqrt(self):
        """Square root which can be called by NumPy"""
        return self ** 0.5

    def log(self):
        """Log which can be called by NumPy"""
        return self.log_value

    @classmethod
    def sum(cls, values):
        """Carry out LogSumExp"""
        log_values = [cls._log(value) for value in values]
        if not log_values:
            return Probability(0)

        max_log_value = max(log_values)
        # Check if all values are zero
        if max_log_value == float('-inf'):
            return Probability(0)

        value_sum = sum(exp(log_value - max_log_value)
                        for log_value in log_values)

        return Probability(cls._log(value_sum) + max_log_value, log_value=True)



# -*- coding: utf-8 -*-
"""Feeder classes are for manipulating data going into the tracker.

A feeder's primary role is to take detection data from inputs into the
framework, and feed them into the tracking algorithms. These can then
optionally be used to drop detections, deliver detections out of sequence,
synchronise out of sequence detections, etc. """
from .base import Feeder

__all__ = ['Feeder']



# -*- coding: utf-8 -*-
import datetime

import pytest
import numpy as np

from ...types.state import GaussianState, State
from ..simple import (
    SingleTargetGroundTruthSimulator, MultiTargetGroundTruthSimulator)


@pytest.fixture(params=[datetime.timedelta(seconds=1),
                        datetime.timedelta(seconds=10),
                        datetime.timedelta(minutes=1)])
def timestep(request):
    return request.param


def test_single_target_ground_truth_simulator(transition_model, timestep):
    initial_state = State(np.array([[1]]), timestamp=datetime.datetime.now())
    simulator = SingleTargetGroundTruthSimulator(transition_model,
                                                 initial_state, timestep)

    for step, (time, groundtruth_paths) in \
            enumerate(simulator.groundtruth_paths_gen()):
        # Check single ground truth track
        assert len(groundtruth_paths) == 1

        # Check length of path is equal to number of steps
        gt_path = groundtruth_paths.pop()
        assert len(gt_path) == step + 1

        # Check time is now + steps
        timedelta = simulator.timestep * step
        assert gt_path[-1].timestamp == initial_state.timestamp + timedelta

        # Check ground truth object has moved
        assert gt_path[-1].state_vector == initial_state.state_vector +\
            timedelta.total_seconds()

    # Check that the number of steps is equal to the simulation steps
    assert step + 1 == simulator.number_steps


def test_multitarget_ground_truth_simulator(transition_model, timestep):
    initial_state = GaussianState(np.array([[1]]), np.array([[0]]),
                                  timestamp=datetime.datetime.now())
    simulator = MultiTargetGroundTruthSimulator(transition_model,
                                                initial_state, timestep)

    total_paths = set()
    for step, (time, groundtruth_paths) in\
            enumerate(simulator.groundtruth_paths_gen()):
        total_paths |= groundtruth_paths

        # Check time is now + steps
        assert time == initial_state.timestamp + simulator.timestep * step

    # Check number of steps is equal to simulation steps
    assert step + 1 == simulator.number_steps

    # Check that there are multiple ground truth paths
    assert len(total_paths) > 1

    # Check that ground truth paths die
    assert len(groundtruth_paths) < len(total_paths)

    # Check that ground truth paths vary in length
    assert len({len(path) for path in total_paths}) > 1



# coding: utf-8
import datetime

import numpy as np

from ...models.transition.linear import ConstantVelocity
from ...predictor.particle import ParticlePredictor
from ...types.particle import Particle
from ...types.prediction import ParticleStatePrediction
from ...types.state import ParticleState


def test_particle():
    # Initialise a transition model
    cv = ConstantVelocity(noise_diff_coeff=0)

    # Define time related variables
    timestamp = datetime.datetime.now()
    timediff = 2  # 2sec
    new_timestamp = timestamp + datetime.timedelta(seconds=timediff)
    time_interval = new_timestamp - timestamp

    # Define prior state
    prior_particles = [Particle(np.array([[10], [10]]),
                                1 / 9),
                       Particle(np.array([[10], [20]]),
                                1 / 9),
                       Particle(np.array([[10], [30]]),
                                1 / 9),
                       Particle(np.array([[20], [10]]),
                                1 / 9),
                       Particle(np.array([[20], [20]]),
                                1 / 9),
                       Particle(np.array([[20], [30]]),
                                1 / 9),
                       Particle(np.array([[30], [10]]),
                                1 / 9),
                       Particle(np.array([[30], [20]]),
                                1 / 9),
                       Particle(np.array([[30], [30]]),
                                1 / 9),
                       ]
    prior = ParticleState(prior_particles, timestamp=timestamp)

    eval_particles = [Particle(cv.matrix(timestamp=new_timestamp,
                                         time_interval=time_interval)
                               @ particle.state_vector,
                               1 / 9)
                      for particle in prior_particles]
    eval_mean = np.mean(np.hstack([i.state_vector for i in eval_particles]),
                        axis=1).reshape(2, 1)

    eval_prediction = ParticleStatePrediction(eval_particles, new_timestamp)

    predictor = ParticlePredictor(transition_model=cv)

    prediction = predictor.predict(prior, timestamp=new_timestamp)

    assert np.allclose(prediction.mean, eval_mean)
    assert prediction.timestamp == new_timestamp
    assert np.all([eval_prediction.particles[i].state_vector ==
                   prediction.particles[i].state_vector for i in range(9)])
    assert np.all([prediction.particles[i].weight == 1 / 9 for i in range(9)])



# -*- coding: utf-8 -*-
from collections.abc import Sequence

from stonesoup.base import Base


def _headings(heading, lines):
    try:
        index = lines.index(heading) + 2
    except ValueError:
        # No placeholder found, so extend.
        # Numpydoc will ignore empty list: no need to check for contents
        if lines and not lines[-1] == "":
            lines.append("")
        lines.extend([heading, "-" * len(heading)])
        index = len(lines)

    return index


def declarative_class(app, what, name, obj, options, lines):
    """Add declared properties to Parameters list for numpydoc"""
    if what == "class" and issubclass(obj, Base):
        param_index = _headings("Parameters", lines)
        attr_index = _headings("Attributes", lines)
        for name, property_ in obj.properties.items():
            is_sequence = isinstance(property_.cls, Sequence)
            if is_sequence:
                cls = property_.cls[0]
            else:
                cls = property_.cls
            class_name = "{}.{}".format(
                cls.__module__, cls.__name__)
            # To shorten names for builtins and also stonesoup components
            tild = class_name.split(".")[0] in ("stonesoup", "builtins")
            # To add optional if default value is defined.
            is_optional = property_.default is not property_.empty
            doc_type = "{}:class:`{}{}`{}".format(
                is_sequence and "sequence of " or "",
                tild and "~" or "",
                class_name,
                is_optional and ", optional" or "",
            )

            new_lines = "{} : {}\n    {}".format(
                name, doc_type, property_.doc or "").split("\n")
            lines[param_index:param_index] = new_lines
            param_index += len(new_lines)
            attr_index += len(new_lines)
            lines.insert(attr_index, name)
            attr_index += 1


def setup(app):
    app.connect('autodoc-process-docstring', declarative_class)



# -*- coding: utf-8 -*-
"""Base classes for Stone Soup feeder"""
from ..base import Property
from ..reader import DetectionReader


class Feeder(DetectionReader):
    """Feeder base class

    Feeder consumes and outputs :class:`.Detection` data and can be used to
    modify the sequence, duplicate or drop data.
    """

    detector = Property(DetectionReader, doc="Source of detections")



# -*- coding: utf-8 -*-
import datetime

import pytest
import numpy as np

from ..neighbour import NearestNeighbour, GlobalNearestNeighbour
from ...types.detection import Detection
from ...types.state import GaussianState
from ...types.track import Track


@pytest.fixture(params=[NearestNeighbour, GlobalNearestNeighbour])
def associator(request, hypothesiser):
    return request.param(hypothesiser)


def test_nearest_neighbour(associator):

    timestamp = datetime.datetime.now()
    t1 = Track([GaussianState(np.array([[0]]), np.array([[1]]), timestamp)])
    t2 = Track([GaussianState(np.array([[3]]), np.array([[1]]), timestamp)])
    d1 = Detection(np.array([[2]]))
    d2 = Detection(np.array([[5]]))

    tracks = {t1, t2}
    detections = {d1, d2}

    associations = associator.associate(tracks, detections, timestamp)

    # There should be 2 associations
    assert len(associations) == 2

    # Each track should associate with a unique detection
    associated_measurements = [hypothesis.measurement
                               for hypothesis in associations.values()
                               if hypothesis.measurement]
    assert len(associated_measurements) == len(set(associated_measurements))


def test_missed_detection_nearest_neighbour(associator):

    timestamp = datetime.datetime.now()
    t1 = Track([GaussianState(np.array([[0]]), np.array([[1]]), timestamp)])
    t2 = Track([GaussianState(np.array([[3]]), np.array([[1]]), timestamp)])
    d1 = Detection(np.array([[20]]))

    tracks = {t1, t2}
    detections = {d1}

    associations = associator.associate(tracks, detections, timestamp)

    # Best hypothesis should be missed detection hypothesis
    assert all(not hypothesis.measurement
               for hypothesis in associations.values())



# -*- coding: utf-8 -*-
from ..base import Base


class Hypothesiser(Base):
    """Hypothesiser base class

    Given a track and set of detections, generate hypothesis of association.
    """

    def hypothesise(self, track, detections, **kwargs):
        """Hypothesise track and detection association

        Parameters
        ----------
        track : Track
            Track which hypotheses will be generated for.
        detections :
            Detections used to generate hypotheses.

        Returns
        -------
        : sequence of :class:`~.Hypothesis`
            Ordered sequence of "best" to "worse" hypothesis.
        """
        raise NotImplementedError



# coding: utf-8
import datetime

import scipy as sp
from scipy.stats import multivariate_normal

from ..linear import ConstantVelocity


def test_cvmodel():
    """ ConstanVelocity Transition Model test """

    # State related variables
    state_vec = sp.array([[3.0], [1.0]])
    old_timestamp = datetime.datetime.now()
    timediff = 1  # 1sec
    new_timestamp = old_timestamp + datetime.timedelta(seconds=timediff)
    time_interval = new_timestamp - old_timestamp

    # Model-related components
    noise_diff_coeff = 0.001  # m/s^2
    F = sp.array([[1, timediff], [0, 1]])
    Q = sp.array([[sp.power(timediff, 3)/3,
                   sp.power(timediff, 2)/2],
                  [sp.power(timediff, 2)/2,
                   timediff]]) * noise_diff_coeff

    # Create and a Constant Velocity model object
    cv = ConstantVelocity(noise_diff_coeff=noise_diff_coeff)

    # Ensure ```cv.transfer_function(time_interval)``` returns F
    assert sp.array_equal(F, cv.matrix(
        timestamp=new_timestamp, time_interval=time_interval))

    # Ensure ```cv.covar(time_interval)``` returns Q
    assert sp.array_equal(Q, cv.covar(
        timestamp=new_timestamp, time_interval=time_interval))

    # Propagate a state vector throught the model
    # (without noise)
    new_state_vec_wo_noise = cv.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=0)
    assert sp.array_equal(new_state_vec_wo_noise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (without noise)
    prob = cv.pdf(new_state_vec_wo_noise,
                  state_vec,
                  timestamp=new_timestamp,
                  time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_wo_noise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throught the model
    # (with internal noise)
    new_state_vec_w_inoise = cv.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval)
    assert not sp.array_equal(new_state_vec_w_inoise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = cv.pdf(new_state_vec_w_inoise,
                  state_vec,
                  timestamp=new_timestamp,
                  time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_inoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throught the model
    # (with external noise)
    noise = cv.rvs(timestamp=new_timestamp, time_interval=time_interval)
    new_state_vec_w_enoise = cv.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=noise)
    assert sp.array_equal(new_state_vec_w_enoise, F@state_vec+noise)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = cv.pdf(new_state_vec_w_enoise, state_vec,
                  timestamp=new_timestamp, time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_enoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)



# -*- coding: utf-8 -*-
from ..base import Base
from ..reader.base import DetectionReader, GroundTruthReader, SensorDataReader


class Simulator(Base):
    """Simulator base class"""


class DetectionSimulator(Simulator, DetectionReader):
    """Detection Simulator base class"""


class GroundTruthSimulator(Simulator, GroundTruthReader):
    """Ground truth simulator"""


class SensorSimulator(Simulator, SensorDataReader):
    """Sensor Simulator base class"""



# -*- coding: utf-8 -*-
"""Provides an ability to generate and load configuration from YAML.

Stone Soup utilises YAML_ for configuration files. The :doc:`stonesoup.base`
feature of components is exploited in order to store the configuration of the
components used for a run.

.. _YAML: http://yaml.org/"""
from abc import ABC, abstractmethod
from io import StringIO


class Configuration:
    pass


class ConfigurationFile(ABC):
    """Base configuration class."""

    @abstractmethod
    def dump(self, data, stream, *args, **kwargs):
        """Dump configuration to a stream."""
        raise NotImplementedError

    def dumps(self, data, *args, **kwargs):
        """Return configuration as a string."""
        stream = StringIO()
        self.dump(data, stream, *args, **kwargs)
        return stream.getvalue()

    @abstractmethod
    def load(self, stream):
        """Load configuration from a stream."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
import datetime

import numpy as np
import pytest

from ...types.detection import Detection
from ...types.hypothesis import SingleHypothesis
from ...types.particle import Particle
from ...types.prediction import (
    GaussianStatePrediction, GaussianMeasurementPrediction,
    StatePrediction, StateMeasurementPrediction,
    ParticleStatePrediction, ParticleMeasurementPrediction)
from ...types.update import (
    StateUpdate, GaussianStateUpdate, ParticleStateUpdate)


def test_stateupdate():
    """ StateUpdate test """

    with pytest.raises(TypeError):
        StateUpdate()

    state_vector = np.array([[-1.8513], [0.9994], [0], [0]]) * 1e4
    timestamp = datetime.datetime.now()
    prediction = StatePrediction(state_vector=np.array([[1], [1],
                                                        [0.5], [0.5]]),
                                 timestamp=timestamp)
    measurement_prediction = StateMeasurementPrediction(
        state_vector=np.array([[2], [2], [1.5], [1.5]]),
        timestamp=timestamp)
    measurement = Detection(state_vector=np.array([[5], [7]]),
                            timestamp=timestamp)

    state_update = StateUpdate(
        state_vector,
        SingleHypothesis(
            prediction=prediction,
            measurement=measurement,
            measurement_prediction=measurement_prediction),
        timestamp=timestamp)

    assert np.array_equal(state_vector, state_update.state_vector)
    assert np.array_equal(prediction, state_update.hypothesis.prediction)
    assert np.array_equal(measurement_prediction,
                          state_update.hypothesis.measurement_prediction)
    assert np.array_equal(measurement, state_update.hypothesis.measurement)
    assert timestamp == state_update.timestamp


def test_gaussianstateupdate():
    """GaussianStateUpdate test"""

    with pytest.raises(TypeError):
        GaussianStateUpdate()

    state_vector = np.array([[-1.8513], [0.9994], [0], [0]]) * 1e4
    covar = np.ones([4, 4])
    timestamp = datetime.datetime.now()

    prediction = GaussianStatePrediction(state_vector=state_vector,
                                         covar=covar,
                                         timestamp=timestamp)
    meas_pred = GaussianMeasurementPrediction(
        state_vector=np.array([[2], [2], [1.5], [1.5]]),
        covar=np.ones([4, 4]) * 2,
        timestamp=timestamp)
    measurement = Detection(state_vector=np.array([[5], [7]]),
                            timestamp=timestamp)
    state_update = GaussianStateUpdate(state_vector,
                                       covar,
                                       SingleHypothesis(
                                           prediction=prediction,
                                           measurement=measurement,
                                           measurement_prediction=meas_pred),
                                       timestamp=timestamp)

    assert np.array_equal(state_vector, state_update.state_vector)
    assert np.array_equal(covar, state_update.covar)
    assert np.array_equal(prediction, state_update.hypothesis.prediction)
    assert np.array_equal(meas_pred,
                          state_update.hypothesis.measurement_prediction)
    assert np.array_equal(measurement, state_update.hypothesis.measurement)
    assert np.array_equal(timestamp, state_update.timestamp)


def test_particlestateupdate():
    """ParticleStateUpdate test"""

    particles = [Particle(np.array([[10], [10]]),
                          1 / 9),
                 Particle(np.array([[10], [20]]),
                          1 / 9),
                 Particle(np.array([[10], [30]]),
                          1 / 9),
                 Particle(np.array([[20], [10]]),
                          1 / 9),
                 Particle(np.array([[20], [20]]),
                          1 / 9),
                 Particle(np.array([[20], [30]]),
                          1 / 9),
                 Particle(np.array([[30], [10]]),
                          1 / 9),
                 Particle(np.array([[30], [20]]),
                          1 / 9),
                 Particle(np.array([[30], [30]]),
                          1 / 9),
                 ]
    timestamp = datetime.datetime.now()
    prediction = ParticleStatePrediction(particles=particles,
                                         timestamp=timestamp)
    meas_pred = ParticleMeasurementPrediction(
        particles=particles,
        timestamp=timestamp)
    measurement = Detection(state_vector=np.array([[5], [7]]),
                            timestamp=timestamp)
    state_update = ParticleStateUpdate(particles,
                                       SingleHypothesis(
                                           prediction=prediction,
                                           measurement=measurement,
                                           measurement_prediction=meas_pred),
                                       timestamp=timestamp)

    eval_mean = np.mean(np.hstack([i.state_vector for i in particles]),
                        axis=1).reshape(2, 1)
    assert np.allclose(eval_mean, state_update.mean)
    assert np.all([particles[i].state_vector ==
                   state_update.particles[i].state_vector for i in range(9)])
    assert np.array_equal(prediction, state_update.hypothesis.prediction)
    assert np.array_equal(meas_pred,
                          state_update.hypothesis.measurement_prediction)
    assert np.array_equal(measurement, state_update.hypothesis.measurement)
    assert np.array_equal(timestamp, state_update.timestamp)



# -*- coding: utf-8 -*-
import pytest
import numpy as np

from ..detection import Detection
from ..multihypothesis import MultipleHypothesis
from ..hypothesis import SingleProbabilityHypothesis, SingleDistanceHypothesis
from ..prediction import GaussianStatePrediction, GaussianMeasurementPrediction
from ..numeric import Probability


def test_multiplehypothesis():
    """MultipleHypothesis type test"""
    timestamp = 1
    prediction = GaussianStatePrediction([[1], [1]],
                                         np.diag([10, 10]),
                                         timestamp=timestamp)
    measurement_prediction = \
        GaussianMeasurementPrediction(np.array([[0.991], [1.062]]),
                                      np.diag([2, 2]),
                                      timestamp=timestamp,
                                      cross_covar=np.diag([10, 10]))
    detection1 = Detection([[2.26], [3.03]], timestamp=timestamp)
    detection2 = Detection([[1.21], [0.95]], timestamp=timestamp)

    probability_hypothesis_1 = SingleProbabilityHypothesis(
        prediction, detection1,
        measurement_prediction=measurement_prediction,
        probability=Probability(0.1))
    probability_hypothesis_2 = SingleProbabilityHypothesis(
        prediction, detection2,
        measurement_prediction=measurement_prediction,
        probability=Probability(0.2))

    multihypothesis = MultipleHypothesis(
        [probability_hypothesis_1, probability_hypothesis_2])

    # check the basic Properties of the MultipleDetectionHypothesis
    assert (hypothesis.prediction is prediction for hypothesis in
            multihypothesis)
    assert (hypothesis.measurement_prediction is measurement_prediction
            for hypothesis in multihypothesis)
    assert len(multihypothesis) == 2
    assert multihypothesis[0] is \
        probability_hypothesis_1
    assert multihypothesis[1] is \
        probability_hypothesis_2
    assert multihypothesis[detection1].measurement is \
        detection1
    assert multihypothesis[prediction].prediction is \
        prediction
    assert probability_hypothesis_1 in multihypothesis
    assert prediction in multihypothesis
    assert detection1 in multihypothesis

    prediction2 = GaussianStatePrediction([[1], [2]], np.diag([10, 10]),
                                          timestamp=timestamp)
    detection3 = Detection([[2.26], [3.82]], timestamp=timestamp)
    assert prediction2 not in multihypothesis
    assert detection3 not in multihypothesis
    assert multihypothesis[prediction2] is None
    assert multihypothesis[detection3] is None

    multihypothesis = MultipleHypothesis(
        [probability_hypothesis_1, probability_hypothesis_2],
        normalise=True, total_weight=1)

    assert sum(hyp.weight for hyp in multihypothesis).log_value < 1e-12


def test_multiplehypothesis_edge_cases():
    """MultipleHypothesis error casses tests"""

    timestamp = 1
    prediction = GaussianStatePrediction([[1], [1]],
                                         np.diag([10, 10]),
                                         timestamp=timestamp)
    measurement_prediction = \
        GaussianMeasurementPrediction(np.array([[0.991], [1.062]]),
                                      np.diag([2, 2]),
                                      timestamp=timestamp,
                                      cross_covar=np.diag([10, 10]))

    probability_hypothesis_1 = SingleProbabilityHypothesis(
        prediction, Detection([[2.26], [3.03]], timestamp=timestamp),
        measurement_prediction=measurement_prediction,
        probability=Probability(0.1))
    probability_hypothesis_2 = SingleProbabilityHypothesis(
        prediction, Detection([[1.21], [0.95]], timestamp=timestamp),
        measurement_prediction=measurement_prediction,
        probability=Probability(0.2))
    distance_hypothesis_1 = SingleDistanceHypothesis(
        prediction, Detection([[1.21], [0.95]]), 0.5,
        measurement_prediction=measurement_prediction)

    # test case where a non-Hypothesis is passed into MultipleHypothesis
    # constructor
    with pytest.raises(ValueError):
        multihypothesis = MultipleHypothesis(
            [probability_hypothesis_1, probability_hypothesis_2, 'a'])

    # test case where a non-ProbabilityHypothesis is in a
    # MultipleHypothesis and "normalise_probabilities() is called
    multihypothesis = MultipleHypothesis(
        [probability_hypothesis_1, probability_hypothesis_2,
         distance_hypothesis_1])
    with pytest.raises(ValueError):
        multihypothesis.normalise_probabilities(1)

    # test case where 'normalise' is called with 'total_weight'
    # not equal to a number
    with pytest.raises(TypeError):
        multihypothesis = MultipleHypothesis(
            [probability_hypothesis_1, probability_hypothesis_2],
            normalise=True, total_weight='a')

    # test the case where "get_missed_detection_probability()" is called on a
    # MultipleHypothesis that has no MissedDetection
    multihypothesis = MultipleHypothesis(
        [probability_hypothesis_1, probability_hypothesis_2])
    assert multihypothesis.get_missed_detection_probability() is None

    # test the case where no SingleHypotheses are passed in
    multihypothesis = MultipleHypothesis()
    assert len(multihypothesis) == 0



# -*- coding: utf-8 -*-
from .base import Hypothesiser
from ..base import Property


class FilteredDetectionsHypothesiser(Hypothesiser):
    """Wrapper for Hypothesisers - filters input data

    Wrapper for any type of hypothesiser - filters the 'detections' before
    they are fed into the hypothesiser.
    """

    hypothesiser = Property(
        Hypothesiser, doc="Hypothesiser that is being wrapped.")
    metadata_filter = Property(
        str, doc="Metadata attribute used to filter which detections "
                 "tracks are valid for association.")
    match_missing = Property(
        bool,
        default=True,
        doc="Match detections with missing metadata. Default 'True'.")

    def hypothesise(self, track, detections, *args, **kwargs):
        """
        Parameters
        ==========
        track : :class:`Track`
            A track that contains the target's state
        detections : list of :class:`Detection`
            Retrieved measurements

        Returns
        =======
        : :class:`MultipleHypothesis`
            A list containing the hypotheses between each prediction-detections
            pair.

        Note:   The specific subclass of :class:`SingleHypothesis` returned
                depends on the :class:`Hypothesiser` used.

        """
        track_metadata = track.metadata.get(self.metadata_filter)

        if (track_metadata is None) and self.match_missing:
            match_detections = detections
        else:
            match_metadata = [track_metadata]
            if self.match_missing:
                match_metadata.append(None)

            match_detections = {
                detection for detection in detections
                if detection.metadata.get(
                        self.metadata_filter) in match_metadata}

        return self.hypothesiser.hypothesise(
            track, match_detections, *args, **kwargs)



# -*- coding: utf-8 -*-
"""Test for updater.particle module"""
import numpy as np
import datetime

from ...models.measurement.linear import LinearGaussian
from ...resampler.particle import SystematicResampler
from ...types.detection import Detection
from ...types.hypothesis import SingleHypothesis
from ...types.particle import Particle
from ...types.prediction import (
    ParticleStatePrediction, ParticleMeasurementPrediction)
from ...updater.particle import ParticleUpdater


def test_particle():
    # Measurement model
    lg = LinearGaussian(ndim_state=2, mapping=[0],
                        noise_covar=np.array([[0.04]]))
    timestamp = datetime.datetime.now()
    particles = [Particle(np.array([[10], [10]]),
                          1 / 9),
                 Particle(np.array([[10], [20]]),
                          1 / 9),
                 Particle(np.array([[10], [30]]),
                          1 / 9),
                 Particle(np.array([[20], [10]]),
                          1 / 9),
                 Particle(np.array([[20], [20]]),
                          1 / 9),
                 Particle(np.array([[20], [30]]),
                          1 / 9),
                 Particle(np.array([[30], [10]]),
                          1 / 9),
                 Particle(np.array([[30], [20]]),
                          1 / 9),
                 Particle(np.array([[30], [30]]),
                          1 / 9),
                 ]

    prediction = ParticleStatePrediction(particles,
                                         timestamp=timestamp)
    measurement = Detection(np.array([[20]]), timestamp=timestamp)
    resampler = SystematicResampler()
    updater = ParticleUpdater(lg, resampler)
    eval_measurement_prediction = ParticleMeasurementPrediction([
                                            Particle(i.state_vector[0], 1 / 9)
                                            for i in particles],
                                            timestamp=timestamp)

    measurement_prediction = updater.get_measurement_prediction(prediction)

    assert np.all([eval_measurement_prediction.particles[i].state_vector ==
                   measurement_prediction.particles[i].state_vector
                   for i in range(9)])
    assert measurement_prediction.timestamp == timestamp

    updated_state = updater.update(SingleHypothesis(
        prediction, measurement, measurement_prediction))

    # Don't know what the particles will exactly be due to randomness so check
    # some obvious properties

    assert np.all(particle.weight == 1 / 9
                  for particle in updated_state.particles)
    assert updated_state.timestamp == timestamp
    assert updated_state.hypothesis.measurement_prediction \
        == measurement_prediction
    assert updated_state.hypothesis.prediction == prediction
    assert updated_state.hypothesis.measurement == measurement
    assert np.all(
        np.isclose(updated_state.state_vector, np.array([[20], [20]])))



import datetime
import numpy as np

from ..manager import SimpleManager
from ..base import MetricGenerator
from ...dataassociator import Associator
from ...types.association import Association
from ...types.detection import Detection
from ...types.groundtruth import GroundTruthPath
from ...types.metric import Metric
from ...types.state import State
from ...types.track import Track


def test_adddata():
    manager = SimpleManager([], [])

    # Check adding data to empty manager
    tracks = [Track(
        states=State(np.array([[1]]), timestamp=datetime.datetime.now()))]
    truths = [GroundTruthPath(
        states=State(np.array([[2]]), timestamp=datetime.datetime.now()))]
    dets = [Detection(np.array([[3]]), timestamp=datetime.datetime.now())]

    manager.add_data([tracks, truths, dets])

    assert manager.tracks == set(tracks)
    assert manager.groundtruth_paths == set(truths)
    assert manager.detections == set(dets)

    # Check not overwriting data (flag is true by default)

    tracks2 = [Track(
        states=State(np.array([[21]]), timestamp=datetime.datetime.now()))]
    truths2 = [GroundTruthPath(
        states=State(np.array([[22]]), timestamp=datetime.datetime.now()))]
    dets2 = [Detection(np.array([[23]]), timestamp=datetime.datetime.now())]

    manager.add_data([tracks2, truths2, dets2], overwrite=False)

    assert manager.tracks == set(tracks + tracks2)
    assert manager.groundtruth_paths == set(truths + truths2)
    assert manager.detections == set(dets + dets2)

    # Check adding additional data including repeated data
    manager = SimpleManager([], [])
    manager.add_data([tracks, truths, dets])
    manager.add_data([tracks + tracks2, truths + truths2, dets + dets2],
                     overwrite=True)

    assert manager.tracks == set(tracks2 + tracks)
    assert manager.groundtruth_paths == set(truths2 + truths)
    assert manager.detections == set(dets2 + dets)


def test_associate_tracks():
    class DummyAssociator(Associator):

        def associate_tracks(self, tracks, truths):

            associations = set()
            for track in tracks:
                for truth in truths:
                    associations.add(Association({track, truth}))
            return associations

    manager = SimpleManager(associator=DummyAssociator(), generators=[])
    tracks = {Track(
        states=State(np.array([[1]]), timestamp=datetime.datetime.now()))}
    truths = {GroundTruthPath(
        states=State(np.array([[2]]), timestamp=datetime.datetime.now()))}
    manager.add_data((tracks, truths))

    manager.associate_tracks()

    assert manager.association_set.pop().objects == tracks | truths


def test_listtimestamps():
    timestamp1 = datetime.datetime.now()
    timestamp2 = timestamp1 + datetime.timedelta(seconds=10)
    manager = SimpleManager(associator=[], generators=[])
    tracks = [Track(
        states=[State(np.array([[1]]), timestamp=timestamp1)])]
    truths = [GroundTruthPath(
        states=[State(np.array([[2]]), timestamp=timestamp2)])]
    manager.add_data((tracks, truths))

    assert manager.list_timestamps() == [timestamp1, timestamp2]


def test_generate_metrics():
    class DummyGenerator(MetricGenerator):

        def compute_metric(self, manager, *args, **kwargs):
            return Metric(title="Test metric",
                          value=5,
                          generator=self)

    generator1 = DummyGenerator()
    generator2 = DummyGenerator()

    manager = SimpleManager(associator=[], generators=[generator1, generator2])

    metrics = manager.generate_metrics()
    metric1 = [i for i in metrics if i.generator == generator1][0]
    metric2 = [i for i in metrics if i.generator == generator2][0]

    assert len(metrics) == 2
    assert metric1.title == "Test metric"
    assert np.array_equal(metric1.value, 5)
    assert np.array_equal(metric1.generator, generator1)
    assert metric2.title == "Test metric"
    assert np.array_equal(metric2.value, 5)
    assert np.array_equal(metric2.generator, generator2)



# -*- coding: utf-8 -*-
import datetime

from ..filter import MetadataReducer


def test_metadata_reducer(detector):
    feeder = MetadataReducer(detector, metadata_field="colour")

    multi_none = False
    for time, detections in feeder.detections_gen():
        all_colours = [detection.metadata.get('colour')
                       for detection in detections]
        if not multi_none:
            multi_none = len(
                [colour for colour in all_colours if colour is None]) > 1

        colours = [colour for colour in all_colours if colour is not None]
        assert len(colours) == len(set(colours))

        assert "red" in colours
        assert "blue" in colours
        if time < datetime.datetime(2019, 4, 1, 14, 0, 2):
            assert "green" in colours
        else:
            assert "green" not in colours

        assert all(time == detection.timestamp for detection in detections)

    assert multi_none



import numpy as np
from numpy import deg2rad
from pytest import approx

from ..functions import (
    jacobian, gm_reduce_single, mod_bearing, mod_elevation)


def test_jacobian():
    """ jacobian function test """

    # State related variables
    state_mean = np.array([[3.0], [1.0]])

    def f(x):
        return np.array([[1, 1], [0, 1]])@x

    jac = jacobian(f, state_mean)
    jac = jac  # Stop flake8 unused warning


def test_jacobian2():
    """ jacobian function test """

    # Sample functions to compute Jacobian on
    def fun(x):
        """ function for testing scalars i.e. scalar input, scalar output"""
        return 2*x**2

    def fun1d(vec):
        """ test function with vector input, scalar output"""
        out = 2*vec[0]+3*vec[1]
        return out

    def fun2d(vec):
        """ test function with 2d input and 2d output"""
        out = np.empty((2, 1))
        out[0] = 2*vec[0]**2 + 3*vec[1]**2
        out[1] = 2*vec[0]+3*vec[1]
        return out
        x = 3
        jac = jacobian(fun, x)
        assert jac == 4*x

    x = np.array([[1], [2]])
    # Tolerance value to use to test if arrays are equal
    tol = 1.0e-5

    jac = jacobian(fun1d, x)
    T = np.array([2.0, 3.0])

    FOM = np.where(np.abs(jac-T) > tol)
    # Check # of array elements bigger than tol
    assert len(FOM[0]) == 0

    jac = jacobian(fun2d, x)
    T = np.array([[4.0*x[0], 6*x[1]],
                  [2, 3]])
    FOM = np.where(np.abs(jac - T) > tol)
    # Check # of array elements bigger than tol
    assert len(FOM[0]) == 0

    return


def test_gm_reduce_single():

    means = np.array([[1, 2], [3, 4], [5, 6]])
    covars = np.array([[[1, 1], [1, 0.7]],
                       [[1.2, 1.4], [1.3, 2]],
                       [[2, 1.4], [1.2, 1.2]]])
    weights = np.array([1, 2, 5])

    mean, covar = gm_reduce_single(means, covars, weights)

    assert np.allclose(mean, np.array([[4], [5]]))
    assert np.allclose(covar, np.array([[5.675, 5.35], [5.2, 5.3375]]))


def test_bearing():
    bearing_in = [10., 170., 190., 260., 280., 350., 705]
    rad_in = deg2rad(bearing_in)

    bearing_out = [10., 170., -170., -100., -80., -10., -15.]
    rad_out = deg2rad(bearing_out)

    for ind, val in enumerate(rad_in):
        assert rad_out[ind] == approx(mod_bearing(val))


def test_elevation():
    elev_in = [10., 80., 110., 170., 190., 260., 280]
    rad_in = deg2rad(elev_in)

    elev_out = [10., 80., 70., 10., -10., -80., -80.]
    rad_out = deg2rad(elev_out)

    for ind, val in enumerate(rad_in):
        assert rad_out[ind] == approx(mod_elevation(val))



"""GOSPA/OSPA tests."""
import datetime

import numpy as np

from ..manager import SimpleManager
from ..ospametric import GOSPAMetric, OSPAMetric
from ...models.measurement.linear import LinearGaussian
from ...types.detection import Detection
from ...types.groundtruth import GroundTruthPath, GroundTruthState
from ...types.state import State
from ...types.track import Track


def test_gospametric_extractstates():
    """Test GOSPA extract states."""
    generator = GOSPAMetric(
        c=10.0,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))
    # Test state extraction
    time_start = datetime.datetime.now()
    detections = [Detection(state_vector=np.array([[i]]), timestamp=time_start)
                  for i in range(5)]
    tracks = {Track(states=[State(state_vector=[[i]],
                                  timestamp=time_start)]) for i in range(5)}
    truths = {GroundTruthPath(states=[GroundTruthState(state_vector=[[i]],
                                                       timestamp=time_start)])
              for i in range(5)}
    det_states = generator.extract_states(detections)
    assert det_states.states == detections
    track_states = generator.extract_states(tracks)
    assert set(track_states) == set(t.states[0] for t in tracks)
    truth_states = generator.extract_states(truths)
    assert set(truth_states) == set(t.states[0] for t in truths)


def test_gospametric_compute_assignments():
    """Test GOSPA assignment algorithm."""
    num_states = 5
    generator = GOSPAMetric(
        c=10.0,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))
    time_now = datetime.datetime.now()
    track_obj = Track([State(state_vector=[[i]], timestamp=time_now)
                      for i in range(num_states)])
    truth_obj = GroundTruthPath([State(state_vector=[[i]], timestamp=time_now)
                                for i in range(num_states)])
    cost_matrix = generator.compute_cost_matrix(track_obj.states,
                                                truth_obj.states)
    neg_cost_matrix = -1. * (cost_matrix)
    meas_to_truth, truth_to_meas, opt_cost =\
        generator.compute_assignments(neg_cost_matrix,
                                      10 * num_states * num_states)

    assert (opt_cost == 0.0)
    assert np.array_equal(meas_to_truth, truth_to_meas)
    assert np.array_equal(meas_to_truth,
                          np.array([[i for i in range(num_states)]]))


def test_gospametric_cost_matrix():
    """Test GOSPA cost matrix. Also indirectly checks compute distance."""
    num_states = 5
    generator = GOSPAMetric(
        c=10.0,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))
    time_now = datetime.datetime.now()
    track_obj = Track([State(state_vector=[[i]], timestamp=time_now)
                      for i in range(num_states)])
    truth_obj = GroundTruthPath([State(state_vector=[[i]], timestamp=time_now)
                                for i in range(num_states)])
    cost_matrix = generator.compute_cost_matrix(track_obj.states,
                                                truth_obj.states)

    tmp_vec = np.arange(num_states)
    tmp_mat = np.zeros([num_states, num_states])
    for n in range(num_states):
        tmp_mat[n, :] = np.roll(tmp_vec, n)

    tmp_upper = np.triu(tmp_mat)
    test_matrix = tmp_upper + tmp_upper.transpose()
    assert np.array_equal(cost_matrix, test_matrix)


def test_gospametric_compute_gospa_metric():
    """Test compute GOSPA metric."""
    num_states = 5
    generator = GOSPAMetric(
        c=10.0,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))
    time_now = datetime.datetime.now()
    track_obj = Track([State(state_vector=[[i]], timestamp=time_now)
                      for i in range(num_states)])
    truth_obj = GroundTruthPath([State(state_vector=[[i]], timestamp=time_now)
                                for i in range(num_states)])
    single_time_metric, assignment_matrix =\
        generator.compute_gospa_metric(track_obj.states,
                                       truth_obj.states)
    gospa_metric = single_time_metric.value
    assert (gospa_metric['distance'] == 0.0)
    assert (gospa_metric['localisation'] == 0.0)
    assert (gospa_metric['missed'] == 0.0)
    assert (gospa_metric['false'] == 0.0)


def test_gospametric_computemetric():
    """Test GOSPA compute metric."""
    generator = GOSPAMetric(
        c=10.0,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))
    time = datetime.datetime.now()
    # Multiple tracks and truths present at two timesteps
    tracks = {Track(states=[State(state_vector=[[i + 0.5]], timestamp=time),
                            State(state_vector=[[i + 1]],
                                  timestamp=time + datetime.timedelta(
                                  seconds=1))])
              for i in range(5)}
    truths = {GroundTruthPath(
        states=[State(state_vector=[[i]], timestamp=time),
                GroundTruthState(state_vector=[[i]],
                                 timestamp=time + datetime.timedelta(
                                     seconds=1))])
              for i in range(5)}

    manager = SimpleManager([generator])
    manager.add_data([tracks, truths])
    main_metric = generator.compute_metric(manager)

    assert main_metric.title == "GOSPA Metrics"
    assert main_metric.time_range.start_timestamp == time
    assert main_metric.time_range.end_timestamp == time + datetime.timedelta(
        seconds=1)
    first_association = [i for i in main_metric.value
                         if i.timestamp == time][0]
    assert first_association.title == "GOSPA Metric"
    assert first_association.timestamp == time
    assert first_association.generator == generator
    # In the following, distance is divided by the cardinality of
    # of the set since GOSPA is not normalized.
    assert first_association.value['distance'] / 5. == 0.5
    second_association = [
        i for i in main_metric.value if
        i.timestamp == time + datetime.timedelta(seconds=1)][0]
    assert second_association.title == "GOSPA Metric"
    assert second_association.timestamp == time + datetime.timedelta(seconds=1)
    assert second_association.generator == generator
    assert second_association.value['distance'] / 5. == 1


def test_ospametric_extractstates():
    """Test OSPA metric extract states."""
    generator = OSPAMetric(
        c=10,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))

    # Test state extraction
    time_start = datetime.datetime.now()
    detections = [Detection(state_vector=np.array([[i]]), timestamp=time_start)
                  for i in range(5)]
    tracks = {Track(states=[State(state_vector=[[i]],
                                  timestamp=time_start)]) for i in range(5)}
    truths = {GroundTruthPath(states=[GroundTruthState(state_vector=[[i]],
                                                       timestamp=time_start)])
              for i in range(5)}

    det_states = generator.extract_states(detections)
    assert det_states.states == detections
    track_states = generator.extract_states(tracks)
    assert set(track_states) == set(t.states[0] for t in tracks)
    truth_states = generator.extract_states(truths)
    assert set(truth_states) == set(t.states[0] for t in truths)


def test_ospametric_computecostmatrix():
    """Test OSPA metric compute cost matrix."""
    generator = OSPAMetric(
        c=10,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))

    time = datetime.datetime.now()
    track = Track(states=[
        State(state_vector=[[i]], timestamp=time)
        for i in range(5)])
    truth = GroundTruthPath(states=[
        State(state_vector=[[i]], timestamp=time)
        for i in range(5)])

    cost_matrix = generator.compute_cost_matrix(track.states, truth.states)

    assert np.array_equal(cost_matrix, np.array([[0., 1., 2., 3., 4.],
                                                 [1., 0., 1., 2., 3.],
                                                 [2., 1., 0., 1., 2.],
                                                 [3., 2., 1., 0., 1.],
                                                 [4., 3., 2., 1., 0.]]))


def test_ospametric_computeospadistance():
    """Test OSPA metric compute OSPA distance."""
    generator = OSPAMetric(
        c=10,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))

    time = datetime.datetime.now()
    track = Track(states=[
        State(state_vector=[[i]], timestamp=time)
        for i in range(5)])
    truth = GroundTruthPath(states=[
        State(state_vector=[[i + 0.5]], timestamp=time)
        for i in range(5)])

    metric = generator.compute_OSPA_distance(track.states, truth.states)

    assert metric.title == "OSPA distance"
    assert metric.value == 0.5
    assert metric.timestamp == time
    assert metric.generator == generator


def test_ospametric_computemetric():
    """Test OSPA compute metric."""
    generator = OSPAMetric(
        c=10,
        p=1,
        measurement_model_truth=LinearGaussian(1, [0], None),
        measurement_model_track=LinearGaussian(1, [0], None))

    time = datetime.datetime.now()
    # Multiple tracks and truths present at two timesteps
    tracks = {Track(states=[State(state_vector=[[i + 0.5]], timestamp=time),
                            State(state_vector=[[i + 1]],
                                  timestamp=time + datetime.timedelta(
                                     seconds=1))])
              for i in range(5)}
    truths = {GroundTruthPath(
        states=[State(state_vector=[[i]], timestamp=time),
                GroundTruthState(state_vector=[[i]],
                                 timestamp=time+datetime.timedelta(
                                     seconds=1))])
              for i in range(5)}

    manager = SimpleManager([generator])
    manager.add_data([tracks, truths])
    main_metric = generator.compute_metric(manager)

    assert main_metric.title == "OSPA distances"
    assert main_metric.time_range.start_timestamp == time
    assert main_metric.time_range.end_timestamp == time + datetime.timedelta(
        seconds=1)
    first_association = [i for i in main_metric.value
                         if i.timestamp == time][0]
    assert first_association.title == "OSPA distance"
    assert first_association.value == 0.5
    assert first_association.timestamp == time
    assert first_association.generator == generator
    second_association = [i for i in main_metric.value if
                          i.timestamp ==
                          time + datetime.timedelta(seconds=1)][
        0]
    assert second_association.title == "OSPA distance"
    assert second_association.value == 1
    assert second_association.timestamp == time + datetime.timedelta(seconds=1)
    assert second_association.generator == generator



# -*- coding: utf-8 -*-
import pytest

from ..base import Base, Property


@pytest.fixture('session')
def base():
    class _TestBase(Base):
        property_a = Property(int)
        property_b = Property(str)
        property_c = Property(int, default=123)
    return _TestBase



# -*- coding: utf-8 -*-
import pytest
import numpy as np
from scipy.stats import multivariate_normal

from ..nonlinear import (
    BearingElevationGaussianToCartesian, RangeBearingGaussianToCartesian,
    RangeBearingElevationGaussianToCartesian)
from ....functions import jacobian as compute_jac
from ....types.angle import Bearing, Elevation


def h2d(state_vector, translation_offset, rotation_offset):

    xyz = [[state_vector[0, 0] - translation_offset[0, 0]],
           [state_vector[1, 0] - translation_offset[1, 0]],
           [0]]

    # Get rotation matrix
    theta_z = - rotation_offset[2, 0]
    cos_z, sin_z = np.cos(theta_z), np.sin(theta_z)
    rot_z = np.array([[cos_z, -sin_z, 0],
                      [sin_z, cos_z, 0],
                      [0, 0, 1]])

    theta_y = - rotation_offset[1, 0]
    cos_y, sin_y = np.cos(theta_y), np.sin(theta_y)
    rot_y = np.array([[cos_y, 0, sin_y],
                      [0, 1, 0],
                      [-sin_y, 0, cos_y]])

    theta_x = - rotation_offset[0, 0]
    cos_x, sin_x = np.cos(theta_x), np.sin(theta_x)
    rot_x = np.array([[1, 0, 0],
                      [0, cos_x, -sin_x],
                      [0, sin_x, cos_x]])

    rotation_matrix = rot_z@rot_y@rot_x

    xyz_rot = rotation_matrix @ xyz
    x = xyz_rot[0, 0]
    y = xyz_rot[1, 0]
    # z = 0  # xyz_rot[2, 0]

    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)

    return np.array([[Bearing(phi)], [rho]])


def h3d(state_vector,  translation_offset, rotation_offset):

    xyz = [[state_vector[0, 0] - translation_offset[0, 0]],
           [state_vector[1, 0] - translation_offset[1, 0]],
           [state_vector[2, 0] - translation_offset[2, 0]]]

    # Get rotation matrix
    theta_z = - rotation_offset[2, 0]
    cos_z, sin_z = np.cos(theta_z), np.sin(theta_z)
    rot_z = np.array([[cos_z, -sin_z, 0],
                      [sin_z, cos_z, 0],
                      [0, 0, 1]])

    theta_y = - rotation_offset[1, 0]
    cos_y, sin_y = np.cos(theta_y), np.sin(theta_y)
    rot_y = np.array([[cos_y, 0, sin_y],
                      [0, 1, 0],
                      [-sin_y, 0, cos_y]])

    theta_x = - rotation_offset[0, 0]
    cos_x, sin_x = np.cos(theta_x), np.sin(theta_x)
    rot_x = np.array([[1, 0, 0],
                      [0, cos_x, -sin_x],
                      [0, sin_x, cos_x]])

    rotation_matrix = rot_z@rot_y@rot_x

    xyz_rot = rotation_matrix @ xyz
    x = xyz_rot[0, 0]
    y = xyz_rot[1, 0]
    z = xyz_rot[2, 0]

    rho = np.sqrt(x**2 + y**2 + z**2)
    phi = np.arctan2(y, x)
    theta = np.arcsin(z/rho)

    return np.array([[Elevation(theta)], [Bearing(phi)], [rho]])


def hbearing(state_vector, translation_offset, rotation_offset):
    xyz = [[state_vector[0, 0] - translation_offset[0, 0]],
           [state_vector[1, 0] - translation_offset[1, 0]],
           [state_vector[2, 0] - translation_offset[2, 0]]]

    # Get rotation matrix
    theta_z = - rotation_offset[2, 0]
    cos_z, sin_z = np.cos(theta_z), np.sin(theta_z)
    rot_z = np.array([[cos_z, -sin_z, 0],
                      [sin_z, cos_z, 0],
                      [0, 0, 1]])

    theta_y = - rotation_offset[1, 0]
    cos_y, sin_y = np.cos(theta_y), np.sin(theta_y)
    rot_y = np.array([[cos_y, 0, sin_y],
                      [0, 1, 0],
                      [-sin_y, 0, cos_y]])

    theta_x = - rotation_offset[0, 0]
    cos_x, sin_x = np.cos(theta_x), np.sin(theta_x)
    rot_x = np.array([[1, 0, 0],
                      [0, cos_x, -sin_x],
                      [0, sin_x, cos_x]])

    rotation_matrix = rot_z@rot_y@rot_x

    xyz_rot = rotation_matrix @ xyz
    x = xyz_rot[0, 0]
    y = xyz_rot[1, 0]
    z = xyz_rot[2, 0]

    rho = np.sqrt(x**2 + y**2 + z**2)
    phi = np.arctan2(y, x)
    theta = np.arcsin(z/rho)

    return np.array([[Elevation(theta)], [Bearing(phi)]])


@pytest.mark.parametrize(
    "h, ModelClass, state_vec, R , mapping,\
     translation_offset, rotation_offset",
    [
        (   # 2D meas, 2D state
            h2d,
            RangeBearingGaussianToCartesian,
            np.array([[0], [1]]),
            np.array([[0.015, 0],
                      [0, 0.1]]),
            np.array([0, 1]),
            np.array([[1], [-1]]),
            np.array([[0.2], [-.5], [1]])

        ),
        (   # 3D meas, 3D state
            h3d,
            RangeBearingElevationGaussianToCartesian,
            np.array([[1], [2], [2]]),
            np.array([[0.05, 0, 0],
                      [0, 0.015, 0],
                      [0, 0, 0.1]]),
            np.array([0, 1, 2]),
            np.array([[0], [0], [0]]),
            np.array([[.2], [3], [-1]])
        ),
        (   # 2D meas, 3D state
            hbearing,
            BearingElevationGaussianToCartesian,
            np.array([[1], [2], [3]]),
            np.array([[0.05, 0],
                      [0, 0.015]]),
            np.array([0, 1, 2]),
            np.array([[0], [0], [0]]),
            np.array([[-3], [0], [np.pi/3]])
        )
    ],
    ids=["standard", "RBE", "BearingsOnly"]
)
def test_models(h, ModelClass, state_vec, R,
                mapping, translation_offset, rotation_offset):
    """ RangeBearingGaussianToCartesian Measurement Model test """

    ndim_state = state_vec.size

    # Create and a measurement model object
    model = ModelClass(ndim_state=ndim_state,
                       mapping=mapping,
                       noise_covar=R,
                       translation_offset=translation_offset,
                       rotation_offset=rotation_offset)

    # Project a state throught the model
    # (without noise)
    meas_pred_wo_noise = model.function(state_vec, noise=0)
    eval_m = h(state_vec, model.translation_offset, model.rotation_offset)
    assert np.array_equal(meas_pred_wo_noise, eval_m)

    # Ensure ```lg.transfer_function()``` returns H
    def fun(x):
        return model.function(x, noise=0)
    H = compute_jac(fun, state_vec)
    assert np.array_equal(H, model.jacobian(state_vec))
    # Check Jacobian has proper dimensions
    assert H.shape == (model.ndim_meas, ndim_state)

    # Ensure ```lg.covar()``` returns R
    assert np.array_equal(R, model.covar())

    # Project a state throught the model
    # (without noise)
    meas_pred_wo_noise = model.function(state_vec, noise=0)
    assert np.array_equal(meas_pred_wo_noise,  h(
        state_vec, model.translation_offset, model.rotation_offset))

    # Evaluate the likelihood of the predicted measurement, given the state
    # (without noise)
    prob = model.pdf(meas_pred_wo_noise, state_vec)
    assert np.array_equal(prob, multivariate_normal.pdf(
        meas_pred_wo_noise.T,
        mean=np.array(h(state_vec,
                        model.translation_offset,
                        model.rotation_offset)).ravel(),
        cov=R).T)

    # Propagate a state vector through the model
    # (with internal noise)
    meas_pred_w_inoise = model.function(state_vec)
    assert not np.array_equal(
        meas_pred_w_inoise,  h(state_vec, model.translation_offset,
                               model.rotation_offset))

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model.pdf(meas_pred_w_inoise, state_vec)
    assert np.array_equal(prob, multivariate_normal.pdf(
        meas_pred_w_inoise.T,
        mean=np.array(h(state_vec,
                        model.translation_offset,
                        model.rotation_offset)).ravel(),
        cov=R).T)

    # Propagate a state vector throught the model
    # (with external noise)
    noise = model.rvs()
    meas_pred_w_enoise = model.function(state_vec,
                                        noise=noise)
    assert np.array_equal(meas_pred_w_enoise,  h(
        state_vec, model.translation_offset, model.rotation_offset)+noise)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model.pdf(meas_pred_w_enoise, state_vec)
    assert np.array_equal(prob, multivariate_normal.pdf(
        meas_pred_w_enoise.T,
        mean=np.array(h(state_vec,
                        model.translation_offset,
                        model.rotation_offset)).ravel(),
        cov=R).T)



# -*- coding: utf-8 -*-
import pymap3d
import pytest
import utm

from ...types.detection import Detection
from ..geo import LongLatToUTMConverter, LLAtoENUConverter, LLAtoNEDConverter


@pytest.fixture()
def detector():
    class Detector:

        @property
        def detections(self):
            return self._detections

        def detections_gen(self):
            for i in range(-3, 4):
                self._detections = {Detection([[i], [50], [5000 + i*10]])}
                yield None, self.detections

    return Detector()


@pytest.mark.parametrize(
    'converter_class,reverse_func',
    [
        (LLAtoENUConverter, pymap3d.enu2geodetic),
        (LLAtoNEDConverter, pymap3d.ned2geodetic),
    ])
def test_lla_reference_converter(detector, converter_class, reverse_func):
    converter = converter_class(detector, reference_point=(0, 50, 5000))

    for i, (time, detections) in zip(
            range(-3, 4), converter.detections_gen()):
        detection = detections.pop()

        assert pytest.approx((50, i, 5000 + i*10), abs=1e-2, rel=1e-3) == \
            reverse_func(*detection.state_vector[:, 0], 50, 0, 5000)


def test_utm_converter(detector):
    converter = LongLatToUTMConverter(detector)

    p_east = float('-inf')
    assert converter.zone_number is None

    for long, (time, detections) in zip(
            range(-3, 4), converter.detections_gen()):
        detection = detections.pop()

        assert converter.zone_number == 30
        assert converter.northern

        assert p_east < detection.state_vector[0]
        p_east = detection.state_vector[0]

        assert pytest.approx((50, long), rel=1e-2, abs=1e-4) == utm.to_latlon(
            *detection.state_vector[0:2, 0], zone_number=30, northern=True)



import pytest
import datetime

from ..metric import Metric, PlottingMetric, SingleTimePlottingMetric, \
    SingleTimeMetric, TimeRangePlottingMetric, TimeRangeMetric
from ..time import TimeRange
from ...metricgenerator import MetricGenerator


def test_metric():
    with pytest.raises(TypeError):
        Metric()

    title = "Test metric"
    value = 5

    class temp_generator(MetricGenerator):

        def compute_metric(self, manager):
            return 5

    generator = temp_generator()

    temp_metric = Metric(title=title,
                         value=value,
                         generator=generator)

    assert temp_metric.title == title
    assert temp_metric.value == value
    assert temp_metric.generator == generator


def test_plottingmetric():
    with pytest.raises(TypeError):
        PlottingMetric()

    title = "Test metric"
    value = 5

    class temp_generator(MetricGenerator):

        def compute_metric(self, manager):
            return 5

    generator = temp_generator()
    temp_metric = PlottingMetric(title=title,
                                 value=value,
                                 generator=generator)

    assert temp_metric.title == title
    assert temp_metric.value == value
    assert temp_metric.generator == generator


def test_singletimemetric():
    with pytest.raises(TypeError):
        SingleTimeMetric()

    title = "Test metric"
    value = 5
    timestamp = datetime.datetime.now()

    class temp_generator(MetricGenerator):

        def compute_metric(self, manager):
            return 5

    generator = temp_generator()
    temp_metric = SingleTimeMetric(title=title,
                                   value=value,
                                   timestamp=timestamp,
                                   generator=generator)

    assert temp_metric.title == title
    assert temp_metric.value == value
    assert temp_metric.timestamp == timestamp
    assert temp_metric.generator == generator


def test_timerangemetric():
    with pytest.raises(TypeError):
        TimeRangeMetric()

    title = "Test metric"
    value = 5
    timestamp1 = datetime.datetime.now()
    timestamp2 = timestamp1 + datetime.timedelta(seconds=10)
    time_range = TimeRange(start_timestamp=timestamp1,
                           end_timestamp=timestamp2)

    class temp_generator(MetricGenerator):

        def compute_metric(self, manager):
            return 5

    generator = temp_generator()
    temp_metric = TimeRangeMetric(title=title,
                                  value=value,
                                  time_range=time_range,
                                  generator=generator)

    assert temp_metric.title == title
    assert temp_metric.value == value
    assert temp_metric.time_range == time_range
    assert temp_metric.generator == generator


def test_timerangeplottingmetric():
    with pytest.raises(TypeError):
        TimeRangePlottingMetric()

    title = "Test metric"
    value = 5
    timestamp1 = datetime.datetime.now()
    timestamp2 = timestamp1 + datetime.timedelta(seconds=10)
    time_range = TimeRange(start_timestamp=timestamp1,
                           end_timestamp=timestamp2)

    class temp_generator(MetricGenerator):

        def compute_metric(self, manager):
            return 5

    generator = temp_generator()
    temp_metric = TimeRangePlottingMetric(title=title,
                                          value=value,
                                          time_range=time_range,
                                          generator=generator)

    assert temp_metric.title == title
    assert temp_metric.value == value
    assert temp_metric.time_range == time_range
    assert temp_metric.generator == generator


def test_single_timeplottingmetric():
    with pytest.raises(TypeError):
        SingleTimePlottingMetric()

    title = "Test metric"
    value = 5
    timestamp = datetime.datetime.now()

    class temp_generator(MetricGenerator):

        def compute_metric(self, manager):
            return 5

    generator = temp_generator()
    temp_metric = SingleTimePlottingMetric(title=title,
                                           value=value,
                                           timestamp=timestamp,
                                           generator=generator)

    assert temp_metric.title == title
    assert temp_metric.value == value
    assert temp_metric.timestamp == timestamp
    assert temp_metric.generator == generator
